import torch
from typing import Optional

from language_pipes.config import LpConfig
from language_pipes.jobs.job import Job
from language_pipes.jobs.job_data import JobData
from language_pipes.jobs.job_processor import JobContext, JobProcessor, JobState
from language_pipes.util.enums import ComputeStep, JobStatus

class FakeLogger:
    def __init__(self):
        self.messages = []

    def info(self, message):
        self.messages.append(("info", message))

    def error(self, message):
        self.messages.append(("error", message))


class FakeEndModel:
    def __init__(self):
        self.calls = []

    def tokenize(self, job):
        self.calls.append("tokenize")
        job.input_ids = list(range(24))
        job.prompt_tokens = len(job.input_ids)
        job.next_step()

    def compute_embed(self, job, chunk_start=0, chunk_end=-1):
        if job.compute_step == ComputeStep.TOKENIZE:
            self.calls.append("tokenize")
        self.calls.append("compute_embed")
        job.data = JobData()
        job.data.state = torch.zeros((1, 1))
        job.next_step()

    def compute_norm(self, job):
        self.calls.append("compute_norm")
        job.set_norm(job.data.state)

    def compute_head(self, job):
        self.calls.append("compute_head")
        job.set_output(token=0, eos_token=0)

    def set_result(self, job):
        self.calls.append("set_result")
        job.result = "done"


class FakeEndModelContinue(FakeEndModel):
    def compute_head(self, job):
        self.calls.append("compute_head")
        job.set_output(token=1, eos_token=0)


class FakeModel:
    def __init__(self, node_id, start_layer, end_layer, virtual=False, num_hidden_layers=1):
        self.node_id = node_id
        self.start_layer = start_layer
        self.end_layer = end_layer
        self.virtual = virtual
        self.loaded = True
        self.num_hidden_layers = num_hidden_layers

    def process_job(self, job, logger):
        if job.data is None:
            job.data = JobData()
        job.set_layer(torch.zeros((1, 1)), self.end_layer + 1, self.num_hidden_layers)

class TrackingModel(FakeModel):
    def __init__(self, node_id, start_layer, end_layer, virtual=False, num_hidden_layers=1):
        super().__init__(node_id, start_layer, end_layer, virtual=virtual, num_hidden_layers=num_hidden_layers)
        self.processed = False

    def process_job(self, job, logger):
        self.processed = True
        super().process_job(job, logger)

class FakePipe:
    def __init__(self, model):
        self._model = model
        self.sent_jobs = []

    def is_complete(self):
        return True

    def get_layer(self, layer, need_physical=False):
        return self._model

    def send_job(self, job, node_id):
        self.sent_jobs.append((job, node_id))


class FakePipeMulti(FakePipe):
    def __init__(self, local_model, next_model):
        super().__init__(local_model)
        self._local_model = local_model
        self._next_model = next_model

    def get_layer(self, layer, need_physical=False):
        if need_physical:
            return self._local_model
        return self._next_model


class EmptyPipe(FakePipe):
    def get_layer(self, layer, need_physical=False):
        return None


class IncompletePipe(FakePipe):
    def is_complete(self):
        return False


def make_config(node_id="node-a", prefill_chunk_size=2):
    return LpConfig(
        logging_level="INFO",
        app_dir=".",
        model_dir="./models",
        oai_port=None,
        node_id=node_id,
        hosted_models=[],
        max_pipes=1,
        model_validation=False,
        prefill_chunk_size=prefill_chunk_size,
    )

def mock_complete(a):
    pass


def make_job(**kwargs):
    """Helper to create a Job with sensible defaults."""
    defaults = {
        "origin_node_id": "node-a",
        "messages": [],
        "pipe_id": "pipe-1",
        "model_id": "model-1",
        "prefill_chunk_size": 6
    }
    defaults.update(kwargs)
    return Job(**defaults)


def make_processor(job=None, pipe: Optional[FakePipe]=None, end_model=None, config=None, logger=None):
    """Helper to create a JobProcessor with sensible defaults."""
    return JobProcessor(
        JobContext(
            config=config or make_config(),
            logger=logger or FakeLogger(),
            job=job,
            pipe=pipe or FakePipe(FakeModel("node-a", 0, 0)),
            end_model=end_model,
        )
    )



from typing import Dict, List, Optional
from dataclasses import dataclass

from distributed_state_network.objects.endpoint import Endpoint

@dataclass(frozen=True)
class DSNodeConfig:
    node_id: str
    credential_dir: str
    port: int
    network_ip: Optional[str]
    aes_key: Optional[str]
    bootstrap_nodes: List[Endpoint]

    @staticmethod
    def from_dict(data: Dict) -> 'DSNodeConfig':
        return DSNodeConfig(
            data["node_id"], 
            data["credential_dir"] if "credential_dir" in data else "credentials",
            data["port"],
            data["network_ip"] if "network_ip" in data else None, 
            data["aes_key"] if "aes_key" in data else None, 
            [Endpoint.from_json(e) for e in data["bootstrap_nodes"]]
        )

    def to_string(self) -> str:
        lines = []
        
        lines.append("")
        lines.append("=" * 60)
        lines.append("  DSNode Configuration Details")
        lines.append("=" * 60)
        
        # Core settings
        lines.append("")
        lines.append("--- Node Settings ---")
        lines.append(f"  {'Node ID:':<18} {self.node_id}")
        lines.append(f"  {'Credential Dir:':<18} {self.credential_dir}")
        lines.append(f"  {'Port:':<18} {self.port}")
        
        # Network settings
        lines.append("")
        lines.append("--- Network Settings ---")
        if self.network_ip:
            lines.append(f"  {'Network IP:':<18} {self.network_ip}")
        else:
            lines.append("  Network IP:         Not configured")
        
        if self.aes_key:
            # Show truncated key for security
            display_key = self.aes_key[:8] + "..." + self.aes_key[-8:] if len(self.aes_key) > 20 else self.aes_key
            lines.append(f"  {'AES Key:':<18} {display_key}")
        else:
            lines.append("  Network Encryption: Disabled")
        
        # Bootstrap nodes
        lines.append("")
        lines.append(f"--- Bootstrap Nodes ({len(self.bootstrap_nodes)}) ---")
        if self.bootstrap_nodes:
            for i, endpoint in enumerate(self.bootstrap_nodes):
                lines.append(f"  Node #{i+1}:          {endpoint.address}:{endpoint.port}")
        else:
            lines.append("  No bootstrap nodes configured (standalone/first node)")
        
        lines.append("")
        lines.append("=" * 60)
        
        return "\n".join(lines)



import os
import logging
from pathlib import Path
from uuid import uuid4
from logging import Logger
from typing import List, Optional, Callable

import torch

from llm_layer_collector import LlmLayerCollector
from llm_layer_collector.auto.auto_layer import AutoDecoderLayer

from language_pipes.util import clone_model

from language_pipes.modeling.meta_model import MetaModel
from language_pipes.modeling.llm_meta_data import LlmMetadata

from language_pipes.jobs.job import Job
from language_pipes.jobs.job_data import jobDataToComputationState, detachCompState

def compute_layers(job_data, device, layers, cache):
    comp_state = jobDataToComputationState(job_data, device)
    comp_state = detachCompState(comp_state)
    
    with torch.inference_mode():
        for lyr in layers:
            comp_state.state = lyr(comp_state, cache).detach()
    
    return comp_state.state.detach()

class LlmModel:
    model_id: str
    meta_data: LlmMetadata
    process_id: str
    pipe_id: str
    collector: LlmLayerCollector

    node_id: str
    device: str
    virtual: bool
    model_dir: str

    layers: List[AutoDecoderLayer]
    tokenizer: Callable

    start_layer: int
    end_layer: int
    loaded: bool
    num_hidden_layers: int

    def __init__(
            self,
            model_id: str,
            node_id: str,
            pipe_id: str,
            device: str,
            model_dir: str,
            process_id: Optional[str] = None,
            virtual: bool = False
    ):
        self.model_id = model_id
        self.node_id = node_id
        self.pipe_id = pipe_id
        self.loaded = False
        self.virtual = virtual
        self.layers = []
        self.start_layer = -1
        self.end_layer = -1
        self.device = device
        self.model_dir = model_dir

        if not virtual:
            model_path = str(Path(model_dir) / self.model_id)
            if not os.path.exists(model_path):
                clone_model(model_id, model_path)
            self.collector = LlmLayerCollector(
                    model_dir=os.path.join(model_path, 'data'),
                    cache_file=os.path.join(model_path, 'cache.json'),
                    device=device,
                    dtype=torch.float16 
            )
            self.num_hidden_layers = self.collector.config.num_hidden_layers
            self.meta_data = LlmMetadata(model_path)

        if process_id is None:
            self.process_id = str(uuid4())
        else:
            self.process_id = process_id
            
    def load(self):
        if self.end_layer > self.num_hidden_layers:
            self.end_layer = self.num_hidden_layers - 1

        if self.start_layer == -1 or self.end_layer == -1:
            self.layers = []
        else:
            self.layers = self.collector.load_layer_set(self.start_layer, self.end_layer, self.device)
        self.loaded = True
        self.virtual = False

    def print(self, logger: logging.Logger):
        logger.info(f'''
=================================
Loaded Model: {self.model_id}
Pipe ID: {self.pipe_id}
Node: {self.node_id}
Process: {self.process_id}
Start Layer: {self.start_layer}
End Layer: {self.end_layer}
Device: {self.device}
=================================
''')

    def process_job(self, job: Job, logger: Logger):
        job.timing_stats.add_layer_time(self.node_id, job.current_layer, self.end_layer)
        self.compute_layers(job)
        job.timing_stats.set_send_time(logger)

    def compute_layers(
        self, 
        job: Job
    ):
        if job.data is None:
            raise Exception("cannot compute layers without job data")
        
        job.set_layer(
            state=compute_layers(
                job.data,
                self.device, 
                self.layers, 
                job.cache,
            ), 
            layer=self.end_layer + 1, 
            num_hidden_layers=self.num_hidden_layers
        )
    
    def to_meta(self) -> MetaModel:
        return MetaModel(
            process_id=self.process_id,
            start_layer=self.start_layer,
            end_layer=self.end_layer,
            node_id=self.node_id,
            pipe_id=self.pipe_id,
            model_id=self.model_id,
            loaded=self.loaded,
            num_layers=self.num_hidden_layers,
            meta_data=self.meta_data
        )

    def cleanup_tensors(self):
        torch.cuda.empty_cache()
        del self.layers
        torch.cuda.empty_cache()

    @staticmethod
    def from_meta(meta: MetaModel, model_dir: str) -> 'LlmModel':
        model = LlmModel(
            model_id=meta.model_id,
            node_id=meta.node_id,
            pipe_id=meta.pipe_id,
            device='cpu',
            model_dir=model_dir,
            process_id=meta.process_id,
            virtual=True
        )
        model.loaded = meta.loaded
        model.start_layer = meta.start_layer
        model.end_layer = meta.end_layer
        model.meta_data = meta.meta_data
        model.virtual = True

        return model
    
    @staticmethod
    def from_id(model_dir: str, model_id: str, node_id: str, pipe_id: str, device: str) -> 'LlmModel':
        model = LlmModel(
            model_id=model_id, 
            node_id=node_id, 
            pipe_id=pipe_id, 
            device=device, 
            model_dir=model_dir
        )

        model_path = str(Path(model_dir) / model_id)
        model.meta_data = LlmMetadata(model_path)
        return model



import os
import sys
import torch
import unittest

sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', '..', '..', '..', 'src'))
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', '..', '..', '..', 'tests', 'language_pipes', 'unit'))

from language_pipes.jobs.job_data import JobData
from language_pipes.jobs.job_processor import JobState
from language_pipes.util.enums import ComputeStep, JobStatus

from job_processor.util import make_processor, make_job, mock_complete, FakeEndModel, FakeModel, FakePipe, IncompletePipe, EmptyPipe

class TestFullProcessorRun(unittest.TestCase):
    """End-to-end tests for the full processor run cycle."""

    def test_stops_when_pipe_incomplete(self):
        job = make_job()
        job.compute_step = ComputeStep.TOKENIZE
        end_model = FakeEndModel()

        pipe = IncompletePipe(FakeModel("node-a", 0, 0))
        processor = make_processor(job=job, pipe=pipe, end_model=end_model)

        processor.run()

        self.assertEqual(processor.state, JobState.DONE)
        self.assertEqual(end_model.calls, [])

    def test_processes_local_layers_and_completes(self):
        job = make_job(complete=mock_complete)
        end_model = FakeEndModel()
        model = FakeModel("node-a", 0, 0, virtual=False, num_hidden_layers=1)
        pipe = FakePipe(model)

        processor = make_processor(job=job, pipe=pipe, end_model=end_model)

        processor.run()

        self.assertEqual(job.status, JobStatus.COMPLETED)
        self.assertEqual(job.result, "done")
        self.assertIn("compute_head", end_model.calls)

    def test_sends_job_to_virtual_segment(self):
        job = make_job()
        job.compute_step = ComputeStep.LAYER
        job.current_layer = 0
        job.data = JobData()
        job.data.state = torch.zeros((1, 1))

        model = FakeModel("node-b", 0, 0, virtual=True, num_hidden_layers=1)
        pipe = FakePipe(model)
        processor = make_processor(job=job, pipe=pipe, end_model=None)

        processor.run()

        self.assertEqual(len(pipe.sent_jobs), 1)
        _, node_id = pipe.sent_jobs[0]
        self.assertEqual(node_id, "node-b")

    def test_missing_model_stops_processing(self):
        job = make_job()
        job.compute_step = ComputeStep.LAYER
        job.current_layer = 1
        job.data = JobData()
        job.data.state = torch.zeros((1, 1))

        processor = make_processor(
            job=job,
            pipe=EmptyPipe(FakeModel("node-a", 0, 0)),
            end_model=None,
        )

        processor.run()

        self.assertEqual(processor.state, JobState.DONE)

if __name__ == "__main__":
    unittest.main()


import os
import sys
import torch
import unittest

sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', '..', '..', '..', 'src'))
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', '..', '..', '..', 'tests', 'language_pipes', 'unit'))

from language_pipes.jobs.job_data import JobData
from language_pipes.jobs.job_processor import JobState
from language_pipes.util.enums import ComputeStep

from job_processor.util import make_processor, make_job, FakeEndModel

class TestValidatingState(unittest.TestCase):
    """Tests for the _state_validating method."""

    def test_stops_when_job_missing(self):
        processor = make_processor(job=None, end_model=FakeEndModel())
        next_state = processor._state_validating()
        self.assertEqual(next_state, JobState.DONE)

    def test_transitions_to_head_when_prefill_done(self):
        job = make_job()
        job.compute_step = ComputeStep.HEAD
        job.prompt_tokens = 2
        job.data = JobData()
        job.data.state = torch.zeros((1, 1))

        processor = make_processor(job=job, end_model=FakeEndModel())
        next_state = processor._state_validating()

        self.assertEqual(next_state, JobState.HEAD)

    def test_transitions_to_embed_when_prefill_has_more_chunks(self):
        job = make_job()
        job.compute_step = ComputeStep.HEAD
        job.current_token = 0
        job.prompt_tokens = 4
        job.init_chunking(chunk_size=2)
        job.data = JobData()
        job.data.state = torch.zeros((1, 1))

        processor = make_processor(job=job, end_model=FakeEndModel())
        next_state = processor._state_validating()

        self.assertEqual(next_state, JobState.EMBED)

    def test_transitions_to_process_layers_for_local_work(self):
        job = make_job()
        job.compute_step = ComputeStep.LAYER
        job.current_layer = 0
        job.data = JobData()
        job.data.state = torch.zeros((1, 1))

        processor = make_processor(job=job, end_model=None)
        next_state = processor._state_validating()

        self.assertEqual(next_state, JobState.PROCESS_LAYERS)



import os
import torch
from uuid import uuid4
from torch import tensor
from pathlib import Path
from logging import Logger

from transformers.models.auto.tokenization_auto import AutoTokenizer

from llm_layer_collector import LlmLayerCollector
from llm_layer_collector.auto.auto_rms import AutoRMSNorm
from llm_layer_collector.compute import compute_embedding, compute_head

from language_pipes.jobs.job import ComputeStep, Job
from language_pipes.jobs.job_data import computationStateToJobData

from language_pipes.modeling.llm_meta_data import LlmMetadata

from language_pipes.util import clone_model

class EndModel:
    model_id: str
    process_id: str
    device: str
    input_embedding: torch.nn.Embedding
    norm: AutoRMSNorm
    head: torch.nn.Linear
    collector: LlmLayerCollector

    def __init__(self, model_dir: str, model_id: str, device: str):
        self.model_id = model_id
        self.device = device

        self.process_id = str(uuid4())
        model_path = str(Path(model_dir) / self.model_id)
        if not os.path.exists(model_path):
            clone_model(model_id, model_path)
        self.meta_data = LlmMetadata(model_path)
        self.collector = LlmLayerCollector(
            model_dir=os.path.join(model_path, 'data'),
            cache_file=os.path.join(model_path, 'cache.json'),
            device=device,
            dtype=torch.float16
        )
        self.tokenizer = AutoTokenizer.from_pretrained(os.path.join(model_path, 'data'))
    
    def size(self):
        return self.meta_data.embed_size + self.meta_data.head_size

    def load(self):
        self.input_embedding = self.collector.load_input_embedding(self.device)
        self.norm = self.collector.load_norm(self.device)
        self.head = self.collector.load_head(self.device)

    def tokenize(self, job: Job):
        prompt = self.tokenizer.apply_chat_template([m.to_json() for m in job.messages], tokenize=False, chat_template=self.tokenizer.chat_template, add_generation_prompt=True)
        input_tokens = [int(t) for t in self.tokenizer.encode(prompt, return_tensors='pt')[0].numpy()]
        job.input_ids = input_tokens
        job.prompt_tokens = len(input_tokens)
        job.next_step()

    def compute_embed(self, job: Job, logger: Logger, prefill_chunk_size: int):
        if job.compute_step != ComputeStep.EMBED and job.compute_step != ComputeStep.TOKENIZE:
            raise ValueError('Invalid step for embedding')
        if self.input_embedding is None:
            raise RuntimeError("Input Embedding must be loaded before computation")
        
        chunk_start, chunk_end = (0, len(job.input_ids))
        if job.chunking.has_more():
            chunk_start, chunk_end = job.chunking.get_range()
        else:
            chunk_start = chunk_end - 1

        chunk_tokens = job.input_ids[chunk_start:chunk_end]

        comp_state = compute_embedding(
            input_embedder=self.input_embedding, 
            input_ids=tensor([chunk_tokens]), 
            config=self.collector.config, 
            cache=job.cache
        )
        
        job.data = computationStateToJobData(comp_state)
        job.next_step()

    def compute_norm(self, job: Job):
        if job.data is None or job.data.state is None:
            raise RuntimeError("Cannot compute norm without job data")
        norm = self.norm(job.data.state.to(self.device))
        job.set_norm(norm)

    def compute_head(self, job: Job):
        if self.head is None:
            raise RuntimeError("Head must be loaded before computation")
        if job.data is None or job.data.state is None:
            raise RuntimeError("Cannot compute head without job data")
        
        head = compute_head(
            head=self.head, 
            input_ids=job.input_ids, 
            device=self.device, 
            state=job.data.state, 
            top_k=job.top_k,
            top_p=job.top_p,
            min_p=job.min_p,
            temperature=job.temperature,
            presence_penalty=job.presence_penalty
        )

        job.set_output(head, self.collector.config.eos_token_id)
        job.delta = self.tokenizer.decode([job.input_ids[-1]])

    def set_result(self, job: Job):
        res_tokens = job.input_id_tensor()
        if res_tokens is None:
            raise Exception("Cannot decode result tensor: no input ids")
        job.result = self.tokenizer.decode(res_tokens[job.prompt_tokens:])

    def clean_up(self):
        del self.input_embedding
        del self.norm
        del self.head



import os
import sys
import time
import unittest
import requests

sys.path.insert(0, os.path.dirname(__file__))

from base import DSNTestBase, spawn_node


class TestHTTPEndpoints(DSNTestBase):
    def test_http_endpoints_exist(self):
        """All HTTP endpoints should exist (not return 404)"""
        n = spawn_node("http-test-node", "127.0.0.1")
        time.sleep(0.5)
        
        endpoints = ['/hello', '/peers', '/update', '/ping', '/data']
        
        for endpoint in endpoints:
            try:
                response = requests.post(
                    f'http://127.0.0.1:{n.config.port}{endpoint}',
                    data=b'test',
                    timeout=2
                )
                self.assertNotEqual(response.status_code, 404, f"Endpoint {endpoint} not found")
                print(f"Endpoint {endpoint} exists (status: {response.status_code})")
            except Exception as e:
                print(f"Endpoint {endpoint} test failed: {e}")


if __name__ == "__main__":
    unittest.main()



import os
import sys
import unittest

import torch

sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', '..', 'src'))

from language_pipes.jobs.job_data import JobData


class JobDataTests(unittest.TestCase):
    def test_hash_and_validate_state(self):
        job_data = JobData()
        job_data.state = torch.zeros((1, 4))
        job_data.position_ids = torch.tensor([[0, 1, 2, 3]])

        state_hash = job_data.hash_state()

        self.assertTrue(JobData.validate_state(job_data.to_bytes(), state_hash))
        self.assertFalse(job_data.validate_state(job_data.to_bytes(), b"bad-hash"))

    def test_round_trip_bytes(self):
        job_data = JobData()
        job_data.state = torch.ones((2, 2))
        job_data.position_ids = torch.tensor([[0, 1]])
        job_data.cache_position = torch.tensor([1])

        serialized = job_data.to_bytes()
        restored = JobData.from_bytes(serialized)

        self.assertTrue(torch.equal(restored.state, job_data.state))
        self.assertTrue(torch.equal(restored.position_ids, job_data.position_ids))
        self.assertTrue(torch.equal(restored.cache_position, job_data.cache_position))


if __name__ == "__main__":
    unittest.main()



import os
import sys
import time
import shutil
import unittest
import requests

sys.path.insert(0, os.path.dirname(__file__))
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '../../src'))

from distributed_state_network import DSNodeServer, DSNodeConfig
from distributed_state_network.objects.data_packet import DataPacket

from base import DSNTestBase, spawn_node, remove_node

class TestSecurity(DSNTestBase):
    def test_bad_aes_key(self):
        """Invalid AES key should raise an error"""
        try:
            DSNodeServer.start(DSNodeConfig("bad key test", "", 8080, None, "bad.key", []))
            self.fail("Should throw error before this")
        except Exception as e:
            print(e)

    def test_authorization_reject_unencrypted(self):
        """Unencrypted HTTP requests should be rejected"""
        n = spawn_node("node", "127.0.0.1")
        time.sleep(0.5)
        
        try:
            response = requests.post(
                f'http://127.0.0.1:{n.config.port}/ping',
                data=bytes([4]) + b'TEST',
                timeout=2
            )
            self.assertNotEqual(response.status_code, 200)
            print(f"Received status code for bad data: {response.status_code}")
        except Exception as e:
            print(f"Request failed as expected: {e}")

    def test_authorization_accept_encrypted(self):
        """Properly encrypted requests should be accepted"""
        n = spawn_node("node", "127.0.0.1")
        time.sleep(0.5)
        
        encrypted_data = n.node.encrypt_data(bytes([4]) + b'TEST')  # MSG_PING = 4
        response = requests.post(
            f'http://127.0.0.1:{n.config.port}/ping',
            data=encrypted_data,
            headers={'Content-Type': 'application/octet-stream'},
            timeout=2
        )
        
        self.assertEqual(response.status_code, 200)
        decrypted = n.node.decrypt_data(response.content)
        self.assertEqual(decrypted[0], 4)

    def test_version_matching(self):
        """Nodes with mismatched versions should not connect"""
        bootstrap = spawn_node("bootstrap", "127.0.0.1")
        old_version = bootstrap.node.version
        bootstrap.node.version = "bad_version"
        try:
            spawn_node("connector", None, [bootstrap.node.my_con().to_json()])
            self.fail("Should throw error when connecting with version mismatch")
        except Exception as e:
            print(e)
        finally:
            bootstrap.node.version = old_version

    def test_authentication_reset(self):
        """Node with reset credentials should fail to reconnect"""
        bootstrap = spawn_node("bootstrap", "127.0.0.1")
        connector = spawn_node("connector", None, [bootstrap.node.my_con().to_json()])
        connector.stop()
        shutil.rmtree("credentials/connector")
        try:
            connector = spawn_node("connector", None, [bootstrap.node.my_con().to_json()])
            self.fail("Should not be able to authenticate with bootstrap")
        except Exception as e:
            print(e)

    def test_reauthentication(self):
        """Node with preserved credentials should successfully reconnect"""
        if os.path.exists("credentials"):
            shutil.rmtree("credentials")
        bootstrap = spawn_node("bootstrap", "127.0.0.1")
        connector = spawn_node("connector", None, [bootstrap.node.my_con().to_json()])
        connector.stop()
        remove_node(connector)
        connector = spawn_node("connector", None, [bootstrap.node.my_con().to_json()])
        self.assertIn('connector', bootstrap.node.peers())

    def test_data_route_unknown_sender(self):
        """Data packets from unknown senders should be rejected"""
        n = spawn_node("node", "127.0.0.1")
        time.sleep(0.5)

        pkt = DataPacket("unknown", b"", b"payload")
        encrypted = n.node.encrypt_data(bytes([5]) + pkt.to_bytes())  # MSG_DATA = 5

        response = requests.post(
            f'http://127.0.0.1:{n.config.port}/data',
            data=encrypted,
            headers={'Content-Type': 'application/octet-stream'},
            timeout=2
        )
        self.assertEqual(response.status_code, 401)

    def test_data_route_bad_signature(self):
        """Data packets with invalid signatures should be rejected"""
        bootstrap = spawn_node("bootstrap", "127.0.0.1")
        spawn_node("connector", None, [bootstrap.node.my_con().to_json()])

        pkt = DataPacket("connector", b"", b"payload")
        encrypted = bootstrap.node.encrypt_data(bytes([5]) + pkt.to_bytes())

        response = requests.post(
            f'http://127.0.0.1:{bootstrap.config.port}/data',
            data=encrypted,
            headers={'Content-Type': 'application/octet-stream'},
            timeout=2
        )
        self.assertEqual(response.status_code, 401)


if __name__ == "__main__":
    unittest.main()



import os
import sys
sys.path.insert(0, os.path.dirname(__file__))
from base import DSNTestBase, spawn_node

class TestConnectivity(DSNTestBase):
    def test_single_node(self):
        """Single node should see itself in peers"""
        node = spawn_node("one", "127.0.0.1")
        self.assertIn("one", list(node.node.peers()))

    def test_two_nodes(self):
        """Two nodes should discover each other"""
        bootstrap = spawn_node("bootstrap", "127.0.0.1")
        connector = spawn_node("connector", None, [bootstrap.node.my_con().to_json()])
        
        self.assertIn("connector", list(bootstrap.node.peers()))
        self.assertIn("bootstrap", list(bootstrap.node.peers()))
        self.assertIn("connector", list(connector.node.peers()))
        self.assertIn("bootstrap", list(connector.node.peers()))

    def test_many_nodes(self):
        """Many nodes connecting to one bootstrap should all discover each other"""
        bootstrap = spawn_node("bootstrap", "127.0.0.1")
        connectors = [spawn_node(f"node-{i}", None, [bootstrap.node.my_con().to_json()]) for i in range(0, 10)]

        boot_peers = list(bootstrap.node.peers())

        for c in connectors:
            peers = c.node.peers()
            self.assertIn(c.config.node_id, boot_peers)
            self.assertIn("bootstrap", list(peers))
            for i in range(0, 10):
                self.assertIn(f"node-{i}", list(peers))

    def test_multi_bootstrap(self):
        """Multiple bootstrap nodes should propagate peer info across the network"""
        bootstraps = [spawn_node(f"bootstrap-{i}", "127.0.0.1") for i in range(0, 3)]
        for i in range(1, len(bootstraps)):
            bootstraps[i].node.bootstrap(bootstraps[i-1].node.my_con())
        
        connectors = []
        for bs in bootstraps:
            new_connectors = [spawn_node(f"node-{i}", None, [bs.node.my_con().to_json()]) for i in range(len(connectors), len(connectors) + 3)]
            connectors.extend(new_connectors)
        
        for ci in connectors:
            peers = ci.node.peers()
            for cj in connectors:
                self.assertIn(cj.config.node_id, peers)
            for b in bootstraps:
                self.assertIn(b.config.node_id, peers)
        
        for bi in bootstraps:
            peers = bi.node.peers()
            for bj in bootstraps:
                self.assertIn(bj.config.node_id, peers)
            for c in connectors:
                self.assertIn(c.config.node_id, peers)

    def test_connection_from_node(self):
        """Should be able to look up connection info by node ID"""
        n0 = spawn_node("node-0", "127.0.0.1")
        n1 = spawn_node("node-1", None, [n0.node.my_con().to_json()])
        con = n0.node.connection_from_node("node-1")
        self.assertEqual(con.port, n1.config.port)
        try:
            n0.node.connection_from_node("test")
            self.fail("Should throw error if it can't find a matching node")
        except Exception as e:
            print(e)


if __name__ == "__main__":
    import unittest
    unittest.main()



import os
import logging
from typing import Callable, Tuple

from distributed_state_network.util.ecdsa import generate_key_pair

class KeyManager:
    def __init__(
        self,
        key_type: str,
        node_id: str, 
        folder: str,
        public_extension: str,
        private_extension: str,
        gen_keys: Callable[[], Tuple[bytes, bytes]]
    ):
        self.key_type = key_type
        self.node_id = node_id
        self.folder = folder
        self.public_extension = public_extension
        self.private_extension = private_extension
        self.gen_keys = gen_keys

    def write_public(self, node_id: str, cert: bytes):
        if not os.path.exists(self.folder):
            os.mkdir(self.folder)
        if not os.path.exists(f'{self.folder}/{node_id}'):
            os.mkdir(f'{self.folder}/{node_id}')
        with open(f'{self.folder}/{self.node_id}/{node_id}.{self.public_extension}', 'wb') as f:
            f.write(cert)

    def read_public(self, node_id: str) -> bytes:
        if not os.path.exists(f'{self.folder}/{self.node_id}/{node_id}.{self.public_extension}'):
            raise Exception(401, f"Cannot find public ECDSA key for {node_id}")
        with open(f'{self.folder}/{self.node_id}/{node_id}.{self.public_extension}', 'rb') as f:
            return f.read()

    def has_public(self, node_id: str) -> bool:
        return os.path.exists(f'{self.folder}/{self.node_id}/{node_id}.{self.public_extension}')

    def ensure_public(self, node_id: str, public_key: bytes):
        if self.has_public(node_id):
            if not self.verify_public(node_id, public_key):
                raise Exception(401, f"Cannot verify ECDSA key for {node_id}")
        else:
            self.write_public(node_id, public_key)

    def public_path(self, node_id: str):
        return f"{self.folder}/{self.node_id}/{node_id}.{self.public_extension}"

    def verify_public(self, node_id: str, public_key: bytes):
        if not self.has_public(node_id):
            return False
        return public_key == self.read_public(node_id)

    def my_public(self) -> bytes:
        return self.read_public(self.node_id)

    def my_private(self):
        if not os.path.exists(f'{self.folder}/{self.node_id}/{self.node_id}.{self.private_extension}'):
            raise Exception("Private key not found")
        with open(f'{self.folder}/{self.node_id}/{self.node_id}.{self.private_extension}', 'rb') as f:
            return f.read()

    def generate_keys(self):
        if os.path.exists(f'{self.folder}/{self.node_id}/{self.node_id}.{self.private_extension}'):
            return
        logging.getLogger("DSN: " + self.node_id).info(f"Generating {self.key_type} Keys ...")
        cert_bytes, key_bytes = self.gen_keys()
        if not os.path.exists(self.folder):
            os.mkdir(self.folder)
        if not os.path.exists(f'{self.folder}/{self.node_id}'):
            os.mkdir(f'{self.folder}/{self.node_id}')
        with open(f'{self.folder}/{self.node_id}/{self.node_id}.{self.public_extension}', 'wb') as f:
            f.write(cert_bytes)
        with open(f'{self.folder}/{self.node_id}/{self.node_id}.{self.private_extension}', 'wb') as f:
            f.write(key_bytes)

class CredentialManager(KeyManager):
    def __init__(self, folder: str, node_id: str):
        KeyManager.__init__(self, "ECDSA", node_id, folder, "pub", "key", generate_key_pair)



from __future__ import annotations
from logging import Logger
from typing import Dict, Iterable, List, Optional, Tuple

from language_pipes.jobs.job_time import JobTime

def _summary(values: Iterable[float]) -> Optional[dict]:
    values = list(values)
    if not values:
        return None
    return {
        "count": len(values),
        "avg_ms": sum(values) / len(values),
        "min_ms": min(values),
        "max_ms": max(values),
    }

class TimingData:
    job_id: str
    chunk_size: int
    network_ms: List[float]
    network_pairs_ms: Dict[Tuple[str, str], List[float]]
    embed_ms: List[float]
    head_ms: List[float]
    layer_ms: List[float]
    token_ms: List[float]
    all_times: List[List[JobTime]]

    def __init__(self, job_id: str, chunk_size: int = 1):
        self.job_id = job_id
        self.chunk_size = chunk_size
        self.all_times = []
        self.network_ms = []
        self.network_pairs_ms = { }
        self.embed_ms = []
        self.head_ms = []
        self.layer_ms = []
        self.token_ms = []

    def add_times(self, new_times: List[JobTime]) -> None:
        if len(new_times) == 0:
            return
        self.all_times.append(new_times)
        ordered = sorted(new_times, key=lambda lt: lt.receive_time)
        for entry in ordered:
            duration_ms = ((entry.send_time - entry.receive_time) * 1000.0) / self.chunk_size
            if entry.is_embed:
                self.embed_ms.append(duration_ms)
            elif entry.is_head:
                self.head_ms.append(duration_ms)
            else:
                self.layer_ms.append(duration_ms / (entry.end_layer - entry.start_layer))

        for i in range(1, len(ordered)):
            prev = ordered[i - 1]
            current = ordered[i]
            if prev.node_id == current.node_id:
                continue
            latency_ms = (current.receive_time - prev.send_time) * 1000.0
            if latency_ms >= 0:
                self.network_ms.append(latency_ms)
                key = (prev.node_id, current.node_id)
                self.network_pairs_ms.setdefault(key, []).append(latency_ms)

        token_duration_ms = (ordered[-1].send_time - ordered[0].receive_time) * 1000.0
        if token_duration_ms >= 0:
            self.token_ms.append(token_duration_ms)

    def log_summary(self, logger: Logger) -> None:
        def log_line(label: str, stats: Optional[dict]) -> None:
            if stats is None:
                logger.info(f"[Timing] {label}: no samples")
                return
            logger.info(
                f"[Timing] {label}: avg={stats['avg_ms']:.2f}ms "
                f"min={stats['min_ms']:.2f}ms max={stats['max_ms']:.2f}ms "
                f"(n={stats['count']})"
            )

        logger.info(f"[Timing] job={self.job_id[:8]} summary")
        log_line("Network latency", _summary(self.network_ms))
        if self.network_pairs_ms:
            for (source, dest), values in sorted(self.network_pairs_ms.items()):
                stats = _summary(values)
                if stats is None:
                    continue
                logger.info(
                    f"[Timing] Network {source} -> {dest}: avg={stats['avg_ms']:.2f}ms "
                    f"min={stats['min_ms']:.2f}ms max={stats['max_ms']:.2f}ms "
                    f"(n={stats['count']})"
                )
        log_line("Embed", _summary(self.embed_ms))
        log_line("Head", _summary(self.head_ms))
        log_line("Layer", _summary(self.layer_ms))
        log_line("Token", _summary(self.token_ms))

class TimingStats:
    output_times: TimingData
    prefill_times: TimingData
    
    current_times: List[JobTime]

    def __init__(self, job_id: str, prefill_chunk_size: int):
        self.output_times = TimingData(job_id)
        self.prefill_times = TimingData(job_id, prefill_chunk_size)
        self.current_times = []

    def add_timing(self, time: JobTime) -> None:
        self.current_times.append(time)

    def add_embed_time(self, node_id: str) -> None:
        self.add_timing(JobTime(node_id=node_id, is_embed=True))

    def add_layer_time(self, node_id: str, start_layer: int, end_layer: int) -> None:
        self.add_timing(JobTime(node_id=node_id, start_layer=start_layer, end_layer=end_layer))

    def add_head_time(self, node_id: str) -> None:
        self.add_timing(JobTime(node_id=node_id, is_head=True))
    
    def set_send_time(self, logger: Logger) -> None:
        if len(self.current_times) == 0:
            return
        last_time = self.current_times[-1]
        last_time.set_send_time()
        type_str = "LAYER"
        
        if last_time.is_embed:
            type_str = "EMBED"
        
        if last_time.is_head:
            type_str = "HEAD"
        
        elapsed = last_time.send_time - last_time.receive_time
        if type_str == "LAYER":
            elapsed /= last_time.end_layer - last_time.start_layer
        
        logger.info(f"[TIMING] {type_str}: {elapsed*1000.0:.2f}ms")
        
    def receive_network_job(self, times: List[JobTime]) -> None:
        self.current_times = times

    def finalize_token(self, logger: Logger) -> None:
        self.output_times.add_times(self.current_times)
        self.current_times = []
        self.output_times.log_summary(logger)

    def finalize_prefill_chunk(self, logger: Logger) -> None:
        self.prefill_times.add_times(self.current_times)
        self.current_times = []
        self.prefill_times.log_summary(logger)


from distributed_state_network.handler import DSNodeServer
from distributed_state_network.dsnode import DSNode
from distributed_state_network.objects.config import DSNodeConfig
from distributed_state_network.objects.endpoint import Endpoint


import os
import sys
import unittest

sys.path.insert(0, os.path.dirname(__file__))
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '../../src'))

from distributed_state_network import DSNodeServer, DSNodeConfig

from base import DSNTestBase


class TestConfiguration(DSNTestBase):
    def test_config_dict(self):
        """DSNodeConfig.from_dict should parse config correctly"""
        config_dict = {
            "node_id": "node",
            "port": 8000,
            "aes_key": "XXX",
            "bootstrap_nodes": [
                {
                    "address": "127.0.0.1",
                    "port": 8001
                }
            ]
        }

        config = DSNodeConfig.from_dict(config_dict)
        self.assertEqual(config_dict["node_id"], config.node_id)
        self.assertEqual(config_dict["port"], config.port)
        self.assertEqual(config_dict["aes_key"], config.aes_key)
        self.assertTrue(len(config.bootstrap_nodes) > 0)
        self.assertEqual(config_dict["bootstrap_nodes"][0]["address"], config.bootstrap_nodes[0].address)
        self.assertEqual(config_dict["bootstrap_nodes"][0]["port"], config.bootstrap_nodes[0].port)

    def test_aes_key_generation(self):
        """Generated AES key should be 64 characters"""
        key = DSNodeServer.generate_key()
        self.assertEqual(64, len(key))


if __name__ == "__main__":
    unittest.main()



import copy
import torch
from transformers.configuration_utils import PretrainedConfig
from transformers.masking_utils import create_causal_mask, create_sliding_window_causal_mask
from transformers.modeling_attn_mask_utils import AttentionMaskConverter
from transformers.cache_utils import DynamicCache

from llm_layer_collector.auto.auto_rotary import AutoRotaryEmbedding
from llm_layer_collector.state_obj import LLmComputationState

def compute_embedding(
        input_embedder: torch.nn.Embedding,
        input_ids: torch.Tensor,
        config: PretrainedConfig,
        cache: DynamicCache
    ) -> LLmComputationState:
    device = input_embedder.weight.device

    state = LLmComputationState()
    
    past_seen_tokens = cache.get_seq_length()
    
    state.state = input_embedder(input_ids.to(device))

    converter = AttentionMaskConverter(is_causal=True)
    L = input_ids.size()[1]

    attention_mask = converter.to_causal_4d(
        batch_size=1,
        query_length=L,
        key_value_length=past_seen_tokens + L,
        dtype=state.state.dtype,
        device=device
    )
    
    state.cache_position = torch.arange(
        past_seen_tokens, end=past_seen_tokens + L, device=device
    )
    
    state.position_ids = state.cache_position.unsqueeze(0)

    mask_kwargs = {
        "config": config,
        "input_embeds": state.state.detach(),
        "attention_mask": attention_mask,
        "cache_position": state.cache_position,
        "past_key_values": cache,
        "position_ids": state.position_ids
    }
    
    state.causal_mask["full_attention"] = create_causal_mask(**mask_kwargs)
    state.causal_mask["sliding_attention"] = None

    try:
        if "sliding_attention" in config.layer_types or config.sliding_window is not None:
            state.causal_mask["sliding_attention"] = create_sliding_window_causal_mask(**mask_kwargs)
    except AttributeError:
        pass

    if config.model_type == 'gemma3_text':
        state.position_embeddings_global = AutoRotaryEmbedding(config)(state.state.detach(), state.position_ids)
        configCopy = copy.deepcopy(config)
        configCopy.rope_theta = configCopy.rope_local_base_freq
        configCopy.rope_scaling = {"rope_type": "default"}
        
        state.position_embeddings_local = AutoRotaryEmbedding(configCopy)(state.state.detach(), state.position_ids)
    else:
        state.position_embeddings = AutoRotaryEmbedding(config)(state.state.detach(), state.position_ids)
    
    return state

def compute_head(
        head: torch.nn.Linear,
        input_ids: torch.Tensor,
        device: str,
        state: torch.Tensor,
        top_k: int = 1,
        top_p: float = 1,
        min_p: float = 0,
        temperature: float = 1,
        presence_penalty: float = 0
    ) -> torch.Tensor:
    with torch.inference_mode():
        state_on_device = state.detach().clone()[:, -1, :].to(device)
        logits = torch.nn.functional.linear(
            state_on_device, 
            head.weight, 
            head.bias
        ).flatten()
        del state_on_device
        
        # Apply presence penalty to discourage token repetition
        # Subtracts penalty from logits of tokens that have already appeared
        if presence_penalty != 0 and len(input_ids) > 0:
            unique_tokens = set(input_ids)
            for token_id in unique_tokens:
                logits[token_id] -= presence_penalty

        if temperature == 0:
            # Greedy decoding - just pick the top token
            head = int(logits.argmax().item())
        else:
            # Apply temperature scaling to logits before softmax
            # Lower temperature = sharper distribution (more deterministic)
            # Higher temperature = flatter distribution (more random)
            scaled_logits = logits / temperature

            # Apply min_p filtering if specified (min_p > 0)
            # Remove tokens with probability < min_p * max_probability
            if min_p > 0:
                probs = torch.nn.functional.softmax(scaled_logits, dim=0)
                max_prob = probs.max()
                min_prob_threshold = min_p * max_prob
                indices_to_remove = probs < min_prob_threshold
                scaled_logits[indices_to_remove] = float('-inf')

            # Apply top_p (nucleus) filtering if specified (top_p < 1.0)
            # Mask out tokens outside the top-p cumulative probability mass
            if top_p < 1.0:
                sorted_logits, sorted_indices = torch.sort(scaled_logits, descending=True)
                sorted_probs = torch.nn.functional.softmax(sorted_logits, dim=0)
                cumulative_probs = torch.cumsum(sorted_probs, dim=0)
                # Find indices to remove (cumulative prob exceeds top_p)
                sorted_indices_to_remove = cumulative_probs > top_p
                # Shift to keep at least one token
                sorted_indices_to_remove[1:] = sorted_indices_to_remove[:-1].clone()
                sorted_indices_to_remove[0] = False
                # Scatter -inf back to original positions
                indices_to_remove = sorted_indices[sorted_indices_to_remove]
                scaled_logits[indices_to_remove] = float('-inf')

            # Apply top_k filtering if specified (top_k > 0)
            # Mask out non-top-k tokens by setting them to -inf
            if top_k > 0:
                k = min(top_k, scaled_logits.size(0))
                top_k_values, _ = torch.topk(scaled_logits, k)
                threshold = top_k_values[-1]
                scaled_logits = torch.where(scaled_logits < threshold, torch.tensor(float('-inf'), device=scaled_logits.device), scaled_logits)

            probabilities = torch.nn.functional.softmax(scaled_logits, dim=0)
            head = int(torch.multinomial(probabilities, num_samples=1).item())
        
        del logits
        return head


import os
import sys
import unittest
from unittest.mock import patch, MagicMock
from typing import List, Optional

sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', '..', '..', 'src'))

from language_pipes.config import LpConfig, HostedModel
from language_pipes.modeling.model_manager import ModelManager
from language_pipes.modeling.meta_model import MetaModel
from language_pipes.modeling.llm_meta_data import LlmMetadata
from language_pipes.pipes.meta_pipe import MetaPipe
from language_pipes.pipes.router_pipes import RouterPipes


class FakeLogger:
    def __init__(self):
        self.messages = []

    def info(self, message):
        self.messages.append(("info", message))

    def warning(self, message):
        self.messages.append(("warning", message))

    def error(self, message):
        self.messages.append(("error", message))


class FakeConnection:
    def __init__(self, address: str):
        self.address = address


class FakeStateNetworkNode:
    def __init__(self, node_id: str):
        self._node_id = node_id
        self._data = {}
        self._peers = []
        self.logger = None

    def node_id(self):
        return self._node_id

    def read_data(self, node_id: str, key: str):
        return self._data.get((node_id, key))

    def update_data(self, key: str, value: str):
        self._data[(self._node_id, key)] = value

    def peers(self):
        return self._peers

    def connection_from_node(self, node_id: str):
        return FakeConnection("127.0.0.1")

    def add_peer(self, peer_id: str, models=None):
        self._peers.append(peer_id)
        if models is None:
            models = []
        import json
        self._data[(peer_id, "models")] = json.dumps([m.to_json() for m in models])


def make_metadata():
    """Create a fake LlmMetadata for testing."""
    metadata = LlmMetadata()
    metadata.embed_size = 128 * 10**6
    metadata.head_size = 256 * 10**6
    metadata.avg_layer_size = 64 * 10**6
    metadata.embed_hash = "embed_hash"
    metadata.head_hash = "head_hash"
    metadata.layer_hashes = ["l0", "l1", "l2", "l3"]
    return metadata


def make_config(
    node_id="node-a",
    hosted_models: Optional[List[HostedModel]] = None,
    max_pipes=2,
    model_validation=False
):
    """Helper to create a LpConfig with sensible defaults."""
    if hosted_models is None:
        hosted_models = []
    return LpConfig(
        logging_level="INFO",
        app_dir=".",
        model_dir="./models",
        oai_port=None,
        node_id=node_id,
        hosted_models=hosted_models,
        max_pipes=max_pipes,
        model_validation=model_validation,
        prefill_chunk_size=6,
    )


class FakeLlmModel:
    """Mock LlmModel for testing without loading real models."""
    
    def __init__(
        self, 
        model_id: str, 
        node_id: str, 
        pipe_id: str, 
        device: str,
        model_dir: str = "./models",
        num_hidden_layers: int = 4
    ):
        self.model_id = model_id
        self.node_id = node_id
        self.pipe_id = pipe_id
        self.device = device
        self.model_dir = model_dir
        self.start_layer = -1
        self.end_layer = -1
        self.loaded = False
        self.num_hidden_layers = num_hidden_layers
        self.meta_data = make_metadata()
        self.process_id = f"process-{model_id}-{pipe_id}"
        
        # Mock collector with config
        self.collector = MagicMock()
        self.collector.config = MagicMock()
        self.collector.config.num_hidden_layers = num_hidden_layers

    def load(self):
        self.loaded = True

    def cleanup_tensors(self):
        pass

    def print(self, logger):
        logger.info(f"FakeLlmModel: {self.model_id} layers {self.start_layer}-{self.end_layer}")

    def to_meta(self) -> MetaModel:
        return MetaModel(
            process_id=self.process_id,
            start_layer=self.start_layer,
            end_layer=self.end_layer,
            node_id=self.node_id,
            pipe_id=self.pipe_id,
            model_id=self.model_id,
            loaded=self.loaded,
            num_layers=self.num_hidden_layers,
            meta_data=self.meta_data
        )


class FakeEndModel:
    """Mock EndModel for testing without loading real models."""
    
    def __init__(self, model_dir: str, model_id: str, device: str):
        self.model_dir = model_dir
        self.model_id = model_id
        self.device = device
        self.loaded = False

    def load(self):
        self.loaded = True

    def clean_up(self):
        pass


class ModelManagerTests(unittest.TestCase):
    """Tests for ModelManager class."""

    def _create_fake_llm_model(self, model_dir, model_id, node_id, pipe_id, device):
        """Factory function for creating FakeLlmModel instances."""
        return FakeLlmModel(
            model_id=model_id,
            node_id=node_id,
            pipe_id=pipe_id,
            device=device,
            model_dir=model_dir
        )

    @patch('language_pipes.modeling.model_manager.LlmModel')
    @patch('language_pipes.modeling.model_manager.EndModel', FakeEndModel)
    def test_init_with_no_hosted_models(self, mock_llm_model):
        """Test ModelManager initializes correctly with no hosted models."""
        logger = FakeLogger()
        config = make_config()
        node = FakeStateNetworkNode("node-a")
        router = RouterPipes(node)

        manager = ModelManager(logger, config, router)

        self.assertEqual(len(manager.models), 0)
        self.assertEqual(len(manager.end_models), 0)
        self.assertEqual(len(manager.pipes_hosted), 0)

    @patch('language_pipes.modeling.model_manager.LlmModel')
    @patch('language_pipes.modeling.model_manager.EndModel', FakeEndModel)
    def test_stop_clears_models(self, mock_llm_model):
        """Test stop() clears all models and end_models."""
        logger = FakeLogger()
        config = make_config()
        node = FakeStateNetworkNode("node-a")
        router = RouterPipes(node)

        manager = ModelManager(logger, config, router)
        # Manually add some fake models
        fake_model = FakeLlmModel("model-1", "node-a", "pipe-1", "cpu")
        manager.models.append(fake_model)
        fake_end_model = FakeEndModel("./models", "model-1", "cpu")
        manager.end_models.append(fake_end_model)

        manager.stop()

        self.assertEqual(len(manager.models), 0)
        self.assertEqual(len(manager.end_models), 0)

    @patch('language_pipes.modeling.model_manager.LlmModel')
    @patch('language_pipes.modeling.model_manager.EndModel', FakeEndModel)
    def test_get_end_model_returns_matching_model(self, mock_llm_model):
        """Test get_end_model returns the correct EndModel by model_id."""
        logger = FakeLogger()
        config = make_config()
        node = FakeStateNetworkNode("node-a")
        router = RouterPipes(node)

        manager = ModelManager(logger, config, router)
        end_model_1 = FakeEndModel("./models", "model-1", "cpu")
        end_model_2 = FakeEndModel("./models", "model-2", "cpu")
        manager.end_models.append(end_model_1)
        manager.end_models.append(end_model_2)

        result = manager.get_end_model("model-2")

        self.assertIsNotNone(result)
        self.assertEqual(result.model_id, "model-2")

    @patch('language_pipes.modeling.model_manager.LlmModel')
    @patch('language_pipes.modeling.model_manager.EndModel', FakeEndModel)
    def test_get_end_model_returns_none_when_not_found(self, mock_llm_model):
        """Test get_end_model returns None when model_id is not found."""
        logger = FakeLogger()
        config = make_config()
        node = FakeStateNetworkNode("node-a")
        router = RouterPipes(node)

        manager = ModelManager(logger, config, router)
        end_model = FakeEndModel("./models", "model-1", "cpu")
        manager.end_models.append(end_model)

        result = manager.get_end_model("nonexistent-model")

        self.assertIsNone(result)

    @patch('language_pipes.modeling.model_manager.EndModel', FakeEndModel)
    @patch('language_pipes.modeling.model_manager.LlmModel')
    def test_host_model_loads_end_model_when_load_ends_true(self, mock_llm_model_class):
        """Test _host_model loads EndModel when load_ends is True."""
        # Setup mock to return a fake model
        fake_model = FakeLlmModel("model-1", "node-a", "pipe-1", "cpu")
        mock_llm_model_class.from_id.return_value = fake_model
        
        logger = FakeLogger()
        hosted_model = HostedModel(
            id="model-1",
            device="cpu",
            max_memory=1.0,  # 1 GB
            load_ends=True
        )
        config = make_config(hosted_models=[hosted_model])
        node = FakeStateNetworkNode("node-a")
        node.add_peer("node-a", [])
        router = RouterPipes(node)

        manager = ModelManager(logger, config, router)

        # Should have loaded an end model
        self.assertEqual(len(manager.end_models), 1)
        self.assertEqual(manager.end_models[0].model_id, "model-1")
        self.assertTrue(manager.end_models[0].loaded)

    @patch('language_pipes.modeling.model_manager.EndModel', FakeEndModel)
    @patch('language_pipes.modeling.model_manager.LlmModel')
    def test_host_model_does_not_load_end_model_when_load_ends_false(self, mock_llm_model_class):
        """Test _host_model does not load EndModel when load_ends is False."""
        fake_model = FakeLlmModel("model-1", "node-a", "pipe-1", "cpu")
        mock_llm_model_class.from_id.return_value = fake_model

        logger = FakeLogger()
        hosted_model = HostedModel(
            id="model-1",
            device="cpu",
            max_memory=1.0,
            load_ends=False
        )
        config = make_config(hosted_models=[hosted_model])
        node = FakeStateNetworkNode("node-a")
        node.add_peer("node-a", [])
        router = RouterPipes(node)

        manager = ModelManager(logger, config, router)

        self.assertEqual(len(manager.end_models), 0)

    @patch('language_pipes.modeling.model_manager.EndModel', FakeEndModel)
    @patch('language_pipes.modeling.model_manager.LlmModel')
    def test_host_model_creates_new_pipe_when_none_exist(self, mock_llm_model_class):
        """Test _host_model creates a new pipe when no pipes exist for the model."""
        fake_model = FakeLlmModel("model-1", "node-a", "new-pipe", "cpu")
        fake_model.start_layer = 0
        fake_model.end_layer = 3
        mock_llm_model_class.from_id.return_value = fake_model

        logger = FakeLogger()
        hosted_model = HostedModel(
            id="model-1",
            device="cpu",
            max_memory=10.0,  # 10 GB - enough memory to load layers
            load_ends=False
        )
        config = make_config(hosted_models=[hosted_model], max_pipes=1)
        node = FakeStateNetworkNode("node-a")
        node.add_peer("node-a", [])
        router = RouterPipes(node)

        manager = ModelManager(logger, config, router)

        # Should have created at least one pipe
        self.assertGreater(len(manager.pipes_hosted), 0)

    @patch('language_pipes.modeling.model_manager.EndModel', FakeEndModel)
    @patch('language_pipes.modeling.model_manager.LlmModel')
    def test_host_model_respects_max_pipes(self, mock_llm_model_class):
        """Test _host_model respects max_pipes configuration."""
        call_count = [0]
        
        def create_model(*args, **kwargs):
            call_count[0] += 1
            fake_model = FakeLlmModel("model-1", "node-a", f"pipe-{call_count[0]}", "cpu")
            fake_model.start_layer = 0
            fake_model.end_layer = 3
            return fake_model
        
        mock_llm_model_class.from_id.side_effect = create_model

        logger = FakeLogger()
        hosted_model = HostedModel(
            id="model-1",
            device="cpu",
            max_memory=10.0,
            load_ends=False
        )
        config = make_config(hosted_models=[hosted_model], max_pipes=1)
        node = FakeStateNetworkNode("node-a")
        node.add_peer("node-a", [])
        router = RouterPipes(node)

        manager = ModelManager(logger, config, router)

        self.assertEqual(len(manager.pipes_hosted), 1)

    @patch('language_pipes.modeling.model_manager.EndModel', FakeEndModel)
    @patch('language_pipes.modeling.model_manager.LlmModel')
    def test_max_pipes_works_per_model_id(self, mock_llm_model_class):
        """Test _host_model respects max_pipes on a per model basis"""
        call_count = [0]
        
        def create_model(*args, **kwargs):
            call_count[0] += 1
            fake_model = FakeLlmModel("model-1", "node-a", f"pipe-{call_count[0]}", "cpu")
            fake_model.start_layer = 0
            fake_model.end_layer = 3
            return fake_model
        
        mock_llm_model_class.from_id.side_effect = create_model

        logger = FakeLogger()
        hosted_model_1 = HostedModel(
            id="model-1",
            device="cpu",
            max_memory=10.0,
            load_ends=False
        )
        hosted_model_2 = HostedModel(
            id="model-2",
            device="cpu",
            max_memory=10.0,
            load_ends=False
        )
        config = make_config(hosted_models=[hosted_model_1, hosted_model_2], max_pipes=1)
        node = FakeStateNetworkNode("node-a")
        node.add_peer("node-a", [])
        router = RouterPipes(node)

        manager = ModelManager(logger, config, router)

        self.assertEqual(len(manager.pipes_hosted), 2)

    @patch('language_pipes.modeling.model_manager.EndModel', FakeEndModel)
    @patch('language_pipes.modeling.model_manager.LlmModel')
    def test_host_model_adds_to_existing_pipe(self, mock_llm_model_class):
        """Test _host_model adds model segments to existing incomplete pipes."""
        metadata = make_metadata()
        
        # Setup: Create an existing pipe with partial coverage
        existing_model = MetaModel(
            process_id="existing-process",
            start_layer=0,
            end_layer=1,
            loaded=True,
            node_id="node-b",
            pipe_id="existing-pipe",
            model_id="model-1",
            num_layers=4,
            meta_data=metadata
        )

        fake_model = FakeLlmModel("model-1", "node-a", "existing-pipe", "cpu")
        fake_model.start_layer = 2
        fake_model.end_layer = 3
        mock_llm_model_class.from_id.return_value = fake_model

        logger = FakeLogger()
        hosted_model = HostedModel(
            id="model-1",
            device="cpu",
            max_memory=10.0,
            load_ends=False
        )
        config = make_config(hosted_models=[hosted_model], max_pipes=2)
        node = FakeStateNetworkNode("node-a")
        node.add_peer("node-a", [])
        node.add_peer("node-b", [existing_model])
        router = RouterPipes(node)

        manager = ModelManager(logger, config, router)

        self.assertIn("existing-pipe", manager.pipes_hosted["model-1"])

    @patch('language_pipes.modeling.model_manager.EndModel', FakeEndModel)
    @patch('language_pipes.modeling.model_manager.LlmModel')
    def test_get_model_for_pipe_returns_none_when_no_memory(self, mock_llm_model_class):
        """Test _get_model_for_pipe returns None when there's not enough memory."""
        fake_model = FakeLlmModel("model-1", "node-a", "pipe-1", "cpu")
        fake_model.meta_data.avg_layer_size = 500 * 10**6  # 500MB per layer
        mock_llm_model_class.from_id.return_value = fake_model

        logger = FakeLogger()
        config = make_config()
        node = FakeStateNetworkNode("node-a")
        router = RouterPipes(node)

        manager = ModelManager(logger, config, router)
        pipe = MetaPipe("pipe-1", "model-1", [])
        
        # Very small available memory
        available_memory = 10  # 10 bytes - not enough
        remaining_memory, model = manager._get_model_for_pipe("model-1", pipe, "cpu", available_memory)

        self.assertIsNone(model)

    @patch('language_pipes.modeling.model_manager.EndModel', FakeEndModel)
    @patch('language_pipes.modeling.model_manager.LlmModel')
    def test_get_model_for_pipe_calculates_layers_based_on_memory(self, mock_llm_model_class):
        """Test _get_model_for_pipe calculates correct number of layers based on available memory."""
        fake_model = FakeLlmModel("model-1", "node-a", "pipe-1", "cpu", num_hidden_layers=10)
        fake_model.meta_data.avg_layer_size = 100 * 10**6  # 100MB per layer
        mock_llm_model_class.from_id.return_value = fake_model

        logger = FakeLogger()
        config = make_config()
        node = FakeStateNetworkNode("node-a")
        router = RouterPipes(node)

        manager = ModelManager(logger, config, router)
        pipe = MetaPipe("pipe-1", "model-1", [])
        
        # Enough memory for ~5 layers (500MB available, 100MB per layer, -1 for buffer)
        available_memory = 500 * 10**6
        remaining_memory, model = manager._get_model_for_pipe("model-1", pipe, "cpu", available_memory)

        self.assertIsNotNone(model)
        # Should start at layer 0 for empty pipe
        self.assertEqual(model.start_layer, 0)
        self.assertEqual(model.end_layer, 4)

    @patch('language_pipes.modeling.model_manager.EndModel', FakeEndModel)
    @patch('language_pipes.modeling.model_manager.LlmModel')
    @patch('language_pipes.modeling.model_manager.validate_model')
    def test_get_model_for_pipe_validates_when_enabled(self, mock_validate, mock_llm_model_class):
        """Test _get_model_for_pipe validates model metadata when validation is enabled."""
        fake_model = FakeLlmModel("model-1", "node-a", "pipe-1", "cpu")
        mock_llm_model_class.from_id.return_value = fake_model
        mock_validate.return_value = False  # Validation fails

        logger = FakeLogger()
        config = make_config(model_validation=True)
        node = FakeStateNetworkNode("node-a")
        router = RouterPipes(node)

        manager = ModelManager(logger, config, router)
        
        # Create pipe with existing segment to trigger validation
        existing_segment = MetaModel(
            process_id="existing",
            start_layer=0,
            end_layer=1,
            loaded=True,
            node_id="node-b",
            pipe_id="pipe-1",
            model_id="model-1",
            num_layers=4,
            meta_data=make_metadata()
        )
        pipe = MetaPipe("pipe-1", "model-1", [existing_segment])
        
        available_memory = 1000 * 10**6
        remaining_memory, model = manager._get_model_for_pipe("model-1", pipe, "cpu", available_memory)

        # Should return None because validation failed
        self.assertIsNone(model)
        # Should have logged a warning
        warnings = [m for m in logger.messages if m[0] == "warning"]
        self.assertGreater(len(warnings), 0)

    @patch('language_pipes.modeling.model_manager.EndModel', FakeEndModel)
    @patch('language_pipes.modeling.model_manager.LlmModel')
    def test_models_are_loaded_after_init(self, mock_llm_model_class):
        """Test that models are loaded (load() is called) after initialization."""
        fake_model = FakeLlmModel("model-1", "node-a", "pipe-1", "cpu")
        fake_model.start_layer = 0
        fake_model.end_layer = 3
        mock_llm_model_class.from_id.return_value = fake_model

        logger = FakeLogger()
        hosted_model = HostedModel(
            id="model-1",
            device="cpu",
            max_memory=10.0,
            load_ends=False
        )
        config = make_config(hosted_models=[hosted_model], max_pipes=1)
        node = FakeStateNetworkNode("node-a")
        node.add_peer("node-a", [])
        router = RouterPipes(node)

        manager = ModelManager(logger, config, router)

        # All models in the manager should be loaded
        for model in manager.models:
            self.assertTrue(model.loaded)

    @patch('language_pipes.modeling.model_manager.EndModel', FakeEndModel)
    @patch('language_pipes.modeling.model_manager.LlmModel')
    def test_models_added_to_network_router(self, mock_llm_model_class):
        """Test that hosted models are added to the network via router_pipes."""
        fake_model = FakeLlmModel("model-1", "node-a", "pipe-1", "cpu")
        fake_model.start_layer = 0
        fake_model.end_layer = 3
        mock_llm_model_class.from_id.return_value = fake_model

        logger = FakeLogger()
        hosted_model = HostedModel(
            id="model-1",
            device="cpu",
            max_memory=10.0,
            load_ends=False
        )
        config = make_config(hosted_models=[hosted_model], max_pipes=1)
        node = FakeStateNetworkNode("node-a")
        node.add_peer("node-a", [])
        router = RouterPipes(node)

        manager = ModelManager(logger, config, router)

        # Check that the model was added to the network
        import json
        models_data = node.read_data("node-a", "models")
        self.assertIsNotNone(models_data)
        models = json.loads(models_data)
        self.assertGreater(len(models), 0)


class ModelManagerMultipleModelsTests(unittest.TestCase):
    """Tests for ModelManager with multiple hosted models."""

    @patch('language_pipes.modeling.model_manager.EndModel', FakeEndModel)
    @patch('language_pipes.modeling.model_manager.LlmModel')
    def test_multiple_hosted_models(self, mock_llm_model_class):
        """Test ModelManager handles multiple hosted models correctly."""
        call_count = [0]
        
        def create_model(model_dir, model_id, node_id, pipe_id, device):
            call_count[0] += 1
            fake_model = FakeLlmModel(model_id, node_id, pipe_id, device)
            fake_model.start_layer = 0
            fake_model.end_layer = 3
            return fake_model
        
        mock_llm_model_class.from_id.side_effect = create_model

        logger = FakeLogger()
        hosted_models = [
            HostedModel(id="model-1", device="cpu", max_memory=5.0, load_ends=True),
            HostedModel(id="model-2", device="cpu", max_memory=5.0, load_ends=True),
        ]
        config = make_config(hosted_models=hosted_models, max_pipes=4)
        node = FakeStateNetworkNode("node-a")
        node.add_peer("node-a", [])
        router = RouterPipes(node)

        manager = ModelManager(logger, config, router)

        # Should have end models for both
        self.assertEqual(len(manager.end_models), 2)
        model_ids = [m.model_id for m in manager.end_models]
        self.assertIn("model-1", model_ids)
        self.assertIn("model-2", model_ids)


if __name__ == "__main__":
    unittest.main()



from time import time

class ChunkState:
    job_id: str
    current_chunk: int  # Current chunk index being processed (0-based)
    total_chunks: int  # Total chunks for prefill (0 = no chunking needed)
    chunk_size: int  # Size of each chunk
    prompt_length: int  # Total prompt length

    def __init__(self, job_id: str):
        self.job_id = job_id
        self.current_chunk = 0
        self.total_chunks = 0
        self.chunk_size = 0
        self.prompt_length = 0

    def init(self, prompt_length: int, chunk_size: int):
        """Initialize chunking if the prompt exceeds chunk_size."""
        self.prompt_length = prompt_length
        if prompt_length > chunk_size:
            self.chunk_size = chunk_size
            self.total_chunks = (prompt_length + chunk_size - 1) // chunk_size
            self.current_chunk = 0
        else:
            self.total_chunks = 0
            self.current_chunk = 0
            self.chunk_size = 0

    def is_active(self) -> bool:
        return self.total_chunks > 1

    def has_more(self) -> bool:
        return self.is_active() and self.current_chunk < self.total_chunks

    def is_final(self) -> bool:
        return not self.is_active() or self.current_chunk == self.total_chunks - 1

    def get_range(self) -> tuple[int, int]:
        if not self.is_active():
            return (0, self.prompt_length)
        start = self.current_chunk * self.chunk_size
        end = min(start + self.chunk_size, self.prompt_length)
        return (start, end)

    def advance(self):
        self.current_chunk += 1

    def print_start(self, logger):
        if self.is_active():
            logger.info(
                f"prompt_tokens={self.prompt_length}, "
                f"chunks={self.total_chunks}, "
                f"chunk_size={self.chunk_size}"
            )
        else:
            logger.info(f"prompt_tokens={self.prompt_length} (no chunking)")

    def disable(self):
        self.current_chunk = 0
        self.total_chunks = 0
        self.chunk_size = 0



import os
import sys
import time
import unittest

sys.path.insert(0, os.path.dirname(__file__))

from base import DSNTestBase, spawn_node, remove_node

class TestWrapperMethods(DSNTestBase):
    def test_peers_wrapper(self):
        """DSNodeServer.peers() should return peer list"""
        bootstrap = spawn_node("bootstrap", "127.0.0.1")
        peers = bootstrap.peers()
        self.assertIsInstance(peers, list)
        self.assertIn("bootstrap", peers)
        
        connector = spawn_node("connector", None, [bootstrap.node.my_con().to_json()])
        
        bootstrap_peers = bootstrap.peers()
        connector_peers = connector.peers()
        
        self.assertIn("bootstrap", bootstrap_peers)
        self.assertIn("connector", bootstrap_peers)
        self.assertIn("bootstrap", connector_peers)
        self.assertIn("connector", connector_peers)

    def test_read_data_wrapper(self):
        """DSNodeServer.read_data() should read peer data"""
        bootstrap = spawn_node("bootstrap", "127.0.0.1")
        connector = spawn_node("connector", None, [bootstrap.node.my_con().to_json()])
        
        result = bootstrap.read_data("connector", "nonexistent_key")
        self.assertIsNone(result)
        
        connector.node.update_data("test_key", "test_value")
        time.sleep(0.5)
        
        result = bootstrap.read_data("connector", "test_key")
        self.assertEqual("test_value", result)
        
        bootstrap.node.update_data("own_key", "own_value")
        result = bootstrap.read_data("bootstrap", "own_key")
        self.assertEqual("own_value", result)

    def test_update_data_wrapper(self):
        """DSNodeServer.update_data() should update and propagate data"""
        bootstrap = spawn_node("bootstrap", "127.0.0.1")
        connector = spawn_node("connector", None, [bootstrap.node.my_con().to_json()])
        
        bootstrap.update_data("wrapper_key", "wrapper_value")
        time.sleep(0.5)
        
        result = connector.read_data("bootstrap", "wrapper_key")
        self.assertEqual("wrapper_value", result)
        
        result = bootstrap.read_data("bootstrap", "wrapper_key")
        self.assertEqual("wrapper_value", result)
        
        bootstrap.update_data("wrapper_key", "updated_value")
        time.sleep(0.5)
        
        result = connector.read_data("bootstrap", "wrapper_key")
        self.assertEqual("updated_value", result)

    def test_is_shut_down_wrapper(self):
        """DSNodeServer.is_shut_down() should reflect shutdown state"""
        node = spawn_node("shutdown_test", "127.0.0.1")
        
        self.assertFalse(node.is_shut_down())
        
        node.stop()
        self.assertTrue(node.is_shut_down())
        
        remove_node(node)

    def test_node_id_wrapper(self):
        """DSNodeServer.node_id() should return configured node ID"""
        node = spawn_node("test_node_id", "127.0.0.1")
        self.assertEqual("test_node_id", node.node_id())
        
        node2 = spawn_node("another_node", "127.0.0.1")
        self.assertEqual("another_node", node2.node_id())
        
        node3 = spawn_node("node-with-dashes", "127.0.0.1")
        self.assertEqual("node-with-dashes", node3.node_id())

    def test_wrapper_methods_consistency(self):
        """Wrapper methods should be consistent with direct node access"""
        bootstrap = spawn_node("bootstrap", "127.0.0.1")
        connector = spawn_node("connector", None, [bootstrap.node.my_con().to_json()])
        
        self.assertEqual(bootstrap.peers(), bootstrap.node.peers())
        self.assertEqual(connector.peers(), connector.node.peers())
        
        self.assertEqual(bootstrap.node_id(), bootstrap.config.node_id)
        self.assertEqual(connector.node_id(), connector.config.node_id)
        
        self.assertEqual(bootstrap.is_shut_down(), bootstrap.node.shutting_down)
        
        connector.update_data("test", "value")
        time.sleep(0.5)
        
        self.assertEqual(
            bootstrap.read_data("connector", "test"),
            bootstrap.node.read_data("connector", "test")
        )


if __name__ == "__main__":
    unittest.main()



import os
import sys
import torch
import unittest

sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', '..', '..', '..', 'src'))
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', '..', '..', '..', 'tests', 'language_pipes', 'unit'))

from language_pipes.jobs.job_data import JobData
from language_pipes.jobs.job_processor import JobState
from language_pipes.util.enums import ComputeStep

from job_processor.util import make_processor, make_job, FakePipe, FakePipeMulti, FakeModel, TrackingModel, EmptyPipe

class TestProcessLayersState(unittest.TestCase):
    """Tests for the _state_process_layers method."""

    def test_transitions_to_done_when_local_model_missing(self):
        job = make_job()
        job.compute_step = ComputeStep.LAYER
        job.current_layer = 0
        job.data = JobData()
        job.data.state = torch.zeros((1, 1))

        processor = make_processor(
            job=job,
            pipe=EmptyPipe(FakeModel("node-a", 0, 0)),
            end_model=None,
        )

        next_state = processor._state_process_layers()

        self.assertEqual(next_state, JobState.DONE)

    def test_transitions_to_send_for_remote_layer(self):
        job = make_job()
        job.compute_step = ComputeStep.LAYER
        job.current_layer = 0
        job.data = JobData()
        job.data.state = torch.zeros((1, 1))
        job.last_update = 0

        local_model = TrackingModel("node-a", 0, 0, virtual=False, num_hidden_layers=2)
        remote_model = FakeModel("node-b", 1, 1, virtual=True)
        pipe = FakePipeMulti(local_model, remote_model)

        processor = make_processor(job=job, pipe=pipe, end_model=None)

        next_state = processor._state_process_layers()

        self.assertEqual(next_state, JobState.SEND)
        self.assertTrue(local_model.processed)
        self.assertGreater(job.last_update, 0)

    def test_transitions_to_process_layers_for_local_segment(self):
        job = make_job()
        job.compute_step = ComputeStep.LAYER
        job.current_layer = 0
        job.data = JobData()
        job.data.state = torch.zeros((1, 1))
        job.last_update = 0

        local_model = TrackingModel("node-a", 0, 0, virtual=False, num_hidden_layers=2)
        next_model = FakeModel("node-a", 1, 1, virtual=False)
        pipe = FakePipeMulti(local_model, next_model)

        processor = make_processor(job=job, pipe=pipe, end_model=None)

        next_state = processor._state_process_layers()

        self.assertEqual(next_state, JobState.PROCESS_LAYERS)
        self.assertTrue(local_model.processed)
        self.assertGreater(job.last_update, 0)

    def test_transitions_to_head_after_all_layers(self):
        job = make_job()
        job.compute_step = ComputeStep.LAYER
        job.current_layer = 0
        job.data = JobData()
        job.data.state = torch.zeros((1, 1))
        job.last_update = 0

        local_model = TrackingModel("node-a", 0, 1, virtual=False, num_hidden_layers=2)
        pipe = FakePipe(local_model)

        processor = make_processor(job=job, pipe=pipe, end_model=None)

        next_state = processor._state_process_layers()

        self.assertEqual(next_state, JobState.HEAD)
        self.assertTrue(local_model.processed)
        self.assertGreater(job.last_update, 0)

    def test_transitions_to_embed_for_prefill(self):
        job = make_job()
        job.prompt_tokens = 24
        job.init_chunking(6)
        job.compute_step = ComputeStep.LAYER
        job.current_layer = 0
        job.data = JobData()
        job.data.state = torch.zeros((1, 1))
        job.last_update = 0

        local_model = TrackingModel("node-a", 0, 1, virtual=False, num_hidden_layers=2)
        pipe = FakePipe(local_model)

        processor = make_processor(job=job, pipe=pipe, end_model=None)

        next_state = processor._state_process_layers()

        self.assertEqual(next_state, JobState.EMBED)
        self.assertTrue(local_model.processed)
        self.assertGreater(job.last_update, 0)

    def test_transitions_to_head_after_prefill(self):
        job = make_job()
        job.prompt_tokens = 24
        job.init_chunking(6)
        job.current_chunk = 5
        job.compute_step = ComputeStep.LAYER
        job.current_layer = 0
        job.data = JobData()
        job.data.state = torch.zeros((1, 1))
        job.last_update = 0

        local_model = TrackingModel("node-a", 0, 1, virtual=False, num_hidden_layers=2)
        pipe = FakePipe(local_model)

        processor = make_processor(job=job, pipe=pipe, end_model=None)

        next_state = processor._state_process_layers()

        self.assertEqual(next_state, JobState.EMBED)
        self.assertTrue(local_model.processed)
        self.assertGreater(job.last_update, 0)


if __name__ == "__main__":
    unittest.main()


from typing import List, Optional, Protocol, Callable

class StateNetworkNode(Protocol):
    def node_id(self) -> str:
        ...

    def read_data(self, node_id: str, key: str) -> Optional[str]:
        ...

    def update_data(self, key: str, value: str) -> None:
        ...

    def peers(self) -> List[str]:
        ...

    def stop(self) -> None:
        ...

    def is_shut_down(self) -> bool:
        ...

    def send_to_node(self, node_id: str, data: bytes):
        ...

    def set_receive_cb(self, cb: Callable):
        ...
    
    def set_update_cb(self, cb: Callable):
        ...
    
    def set_disconnect_cb(self, cb: Callable):
        ...

    def receive_data(self, data: Callable):
        ...


import os
from pathlib import Path

from language_pipes.config import default_model_dir

def prompt(message: str, default=None, required=False) -> str | None:
    """Prompt user for input with optional default value."""
    if default is not None:
        display = f"{message} [{default}]: "
    else:
        display = f"{message}: "
    
    while True:
        value = input(display).strip()
        if value == "":
            if default is not None:
                return default
            if required:
                print("  This field is required.")
                continue
            return None
        return value


def prompt_int(message: str, default=None, required=False) -> int | None:
    """Prompt user for integer input."""
    while True:
        value = prompt(message, default=str(default) if default else None, required=required)
        if value is None:
            return None
        try:
            return int(value)
        except ValueError:
            print("  Please enter a valid number.")


def prompt_float(message: str, default=None, required=False) -> float | None:
    """Prompt user for float input."""
    while True:
        value = prompt(message, default=str(default) if default else None, required=required)
        if value is None:
            return None
        try:
            return float(value)
        except ValueError:
            print("  Please enter a valid number.")


def prompt_bool(message: str, default=False, required=False) -> bool:
    """Prompt user for yes/no input."""
    default_str = "Y/n" if default else "y/N"
    if required:
        default_str = "Y/N"
    while True:
        value = input(f"{message} [{default_str}]: ").strip().lower()
        if value == "" and not required:
            return default
        if value in ("y", "yes", "true", "1"):
            return True
        if value in ("n", "no", "false", "0"):
            return False
        print("  Please enter 'y' or 'n'.")


def prompt_choice(message: str, choices: list, default=None) -> str | None:
    """Prompt user to select from choices."""
    choices_str = "/".join(choices)
    while True:
        value = prompt(f"{message} ({choices_str})", default=default)
        if value in choices:
            return value
        print(f"  Please choose from: {choices_str}")

def prompt_number_choice(message: str, choices: list, default=None, required=False) -> str:
    print(message)
    for i, choice  in enumerate(choices):
        print(f"[{i}] {choice}")
    if default is not None:
        default_idx = choices.index(default)
    else:
        default_idx = None
    while True:
        try:
            strSelect = prompt("Select number of choice", default=default_idx, required=required)
            if strSelect is None:
                continue
            selection = int(strSelect)
        except KeyboardInterrupt:
            exit()
        except:
            print("Invalid selection")
            continue
        if selection < 0 or selection >= len(choices):
            print("Invalid selection")
        else:
            return choices[selection]
    return None

def prompt_continue():
    prompt("Press Enter to continue...")

def get_available_models(models_dir: str | None = None) -> list:
    """Get list of available models from ~/.cache/language_pipes/models.
    
    Returns a list of model IDs in the format 'org/model' (e.g., 'Qwen/Qwen3-1.7B').
    """
    if models_dir is None:
        models_dir = default_model_dir()
    
    models = []
    if not os.path.exists(models_dir):
        return models
    
    for org in os.listdir(models_dir):
        org_path = os.path.join(models_dir, org)
        if os.path.isdir(org_path):
            for model in os.listdir(org_path):
                model_path = os.path.join(org_path, model)
                if os.path.isdir(model_path):
                    models.append(f"{org}/{model}")
    
    return sorted(models)


def prompt_model_id(message: str, required=False) -> str | None:
    """Prompt user to select a model ID from available models or enter a custom one.
    
    Scans ~/.cache/language_pipes/models for locally available models and presents
    them as numbered choices. Also allows entering a custom HuggingFace model ID.
    """
    available_models = get_available_models()
    
    if available_models:
        print(f"\n{message}")
        print("  Available local models:")
        for i, model in enumerate(available_models):
            print(f"    [{i + 1}] {model}")
        print(f"    [0] Enter custom model ID")
        
        while True:
            try:
                selection_str = prompt("  Select option", required=required)
                if selection_str is None:
                    return None
                selection = int(selection_str)
            except ValueError:
                print("  Please enter a valid number.")
                continue
            except KeyboardInterrupt:
                exit()
            
            if selection == 0:
                return prompt("    Custom model ID", required=required)
            elif 1 <= selection <= len(available_models):
                return available_models[selection - 1]
            else:
                print(f"  Please select a number between 0 and {len(available_models)}.")
    else:
        # No local models available, prompt for custom ID directly
        return prompt(message, required=required)


def get_config_files(config_dir: str):
    return [f.replace(".toml", "") for f in os.listdir(config_dir)]

def select_config(app_dir: str) -> str | None:
    config_dir = str(Path(app_dir) / "configs")
    existing_configs = get_config_files(config_dir)

    if len(existing_configs) > 0:
        load_config = prompt_number_choice("Select Configuration", existing_configs, required=True)
        if load_config is None:
            exit()
        load_config = load_config + ".toml"
    else:
        print("No configs found...")
        return None

    return str(Path(config_dir) / load_config)

def show_banner(text: str):
    print("\n" + "=" * 50)
    print("\t" + text)
    print("=" * 50)


"""Unit tests for language_pipes.

Run all tests with: python -m tests.language_pipes.unit
"""
import os
import sys
import unittest

# Add src to path
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', '..', '..', 'src'))


def load_tests(loader, tests, pattern):
    """Load all tests from this package."""
    package_dir = os.path.dirname(__file__)
    suite = loader.discover(start_dir=package_dir, pattern='test_*.py')
    return suite


def run_tests():
    """Run all unit tests."""
    loader = unittest.TestLoader()
    suite = load_tests(loader, None, None)
    runner = unittest.TextTestRunner(verbosity=2)
    result = runner.run(suite)
    return result.wasSuccessful()


if __name__ == '__main__':
    success = run_tests()
    sys.exit(0 if success else 1)



import os
import sys
import time
import unittest
sys.path.insert(0, os.path.dirname(__file__))
from base import DSNTestBase, spawn_node, remove_node

class TestCallbacks(DSNTestBase):

    def test_disconnect_callback(self):
        """Disconnect callback should be called when a peer disconnects"""
        callback_called = []
        
        def on_disconnect():
            callback_called.append(True)
        
        node1 = spawn_node("node-1", "127.0.0.1", [], on_disconnect)
        node2 = spawn_node("node-2", None, [node1.node.my_con().to_json()])
        
        time.sleep(1)
        node2.stop()
        remove_node(node2)
        time.sleep(10)
        
        self.assertEqual(len(callback_called), 1)

    def test_update_callback(self):
        """Update callback should be called when peer data is updated"""
        callback_called = []
        
        def on_update():
            callback_called.append(True)
        
        node1 = spawn_node("node-1", "127.0.0.1", [], None, on_update)
        node2 = spawn_node("node-2", None, [node1.node.my_con().to_json()])
        
        node2.node.update_data("key", "value")
        time.sleep(1)
        
        self.assertEqual(len(callback_called), 2)

    def test_update_callback_error_handling(self):
        """Update callback errors should not crash the node"""
        def on_update_error():
            raise Exception("This should be captured")
        
        node1 = spawn_node("node-1", "127.0.0.1", [], None, on_update_error)
        node2 = spawn_node("node-2", None, [node1.node.my_con().to_json()])

        node2.node.update_data("key", "value")
        time.sleep(1)
        
        # Should reach this point without crashing
        self.assertTrue(True)


if __name__ == "__main__":
    unittest.main()



from typing import Dict, Optional
from dataclasses import dataclass
from distributed_state_network.util.byte_helper import ByteHelper

@dataclass(frozen=True)
class Endpoint:
    address: Optional[str]
    port: int

    def to_string(self):
        return f"{self.address}:{self.port}"

    def to_bytes(self):
        bts = ByteHelper()
        if self.address is not None:
            bts.write_string(self.address)
        else:
            bts.write_string("")
        bts.write_int(self.port)

        return bts.get_bytes()

    def to_json(self):
        return {
            "address": self.address,
            "port": self.port
        }

    @staticmethod
    def from_bytes(data: bytes):
        bts = ByteHelper(data)
        address = bts.read_string()
        if address == "":
            address = None
        port = bts.read_int()

        return Endpoint(address, port)

    @staticmethod
    def from_json(data: Dict) -> 'Endpoint':
        return Endpoint(data['address'], data['port'])



import os
import sys
import time
import random
import unittest
sys.path.insert(0, os.path.dirname(__file__))
from base import DSNTestBase, spawn_node, remove_node

class TestDisconnect(DSNTestBase):
    def test_reconnect(self):
        """Node should be removed from peers after disconnect"""
        bootstrap = spawn_node("bootstrap", "127.0.0.1")
        connector = spawn_node("connector", None, [bootstrap.node.my_con().to_json()])
        self.assertIn(connector.config.node_id, bootstrap.node.peers())
        connector.stop()
        time.sleep(10)
        self.assertNotIn(connector.config.node_id, bootstrap.node.peers())

    def test_disconnect_and_new_join(self):
        """New nodes should not see disconnected nodes"""
        node1 = spawn_node("node-1", "127.0.0.1")
        node2 = spawn_node("node-2", None, [node1.node.my_con().to_json()])
        node3 = spawn_node("node-3", None, [node1.node.my_con().to_json()])

        time.sleep(1)
        node2.stop()
        remove_node(node2)
        time.sleep(10)
        
        node4 = spawn_node("node-4", None, [node1.node.my_con().to_json()])
        time.sleep(10)
        
        self.assertEqual(["node-1", "node-3", "node-4"], sorted(node1.node.peers()))
        self.assertEqual(["node-1", "node-3", "node-4"], sorted(node3.node.peers()))
        self.assertEqual(["node-1", "node-3", "node-4"], sorted(node4.node.peers()))

    def test_peers_after_disconnect(self):
        """peers() wrapper should reflect disconnected nodes"""
        bootstrap = spawn_node("bootstrap", "127.0.0.1")
        connector = spawn_node("connector", None, [bootstrap.node.my_con().to_json()])
        
        self.assertIn("connector", bootstrap.peers())
        
        connector.stop()
        remove_node(connector)
        
        time.sleep(10)
        
        self.assertNotIn("connector", bootstrap.peers())
        self.assertIn("bootstrap", bootstrap.peers())

    @unittest.skip("Long test")
    def test_churn(self):
        """Network should handle continuous join/leave churn"""
        bootstrap = spawn_node("bootstrap", "127.0.0.1")
        
        stopped = []
        connectors = []
        network_labels = ["bootstrap"]
        for _ in range(5):
            new_connectors = [spawn_node(f"node-{i}", None, [bootstrap.node.my_con().to_json()]) for i in range(len(connectors), len(connectors) + 5)]
            connectors.extend(new_connectors)
            for c in new_connectors:
                network_labels.append(c.config.node_id)
            to_shutdown = random.choice(new_connectors)
            to_shutdown.stop()
            network_labels.remove(to_shutdown.config.node_id)
            stopped.append(to_shutdown)
            time.sleep(6)
            for c in connectors:
                if c.config.node_id not in network_labels:
                    continue
                self.assertEqual(sorted(network_labels), sorted(list(c.node.peers())))

if __name__ == "__main__":
    unittest.main()



from typing import List, Optional

from language_pipes.util.byte_helper import ByteHelper
from language_pipes.util.enums import ComputeStep
from language_pipes.jobs.job_data import JobData
from language_pipes.jobs.job_time import JobTime

class NetworkJob:
    job_id: str
    pipe_id: str
    origin_node_id: str
    current_layer: int
    compute_step: ComputeStep
    data: Optional[JobData]
    data_hash: bytes
    times: List[JobTime]

    def __init__(
        self, 
        job_id: str, 
        pipe_id: str,
        origin_node_id: str,
        current_layer: int,
        data: Optional[JobData],
        data_hash: bytes,
        compute_step: ComputeStep,
        times: List[JobTime] = []
    ):
        self.job_id = job_id
        self.pipe_id = pipe_id
        self.origin_node_id = origin_node_id
        self.current_layer = current_layer
        self.data = data
        self.data_hash = data_hash
        self.compute_step = compute_step
        self.times = times

    def to_bytes(self):
        bts = ByteHelper()
        bts.write_string(self.job_id)
        bts.write_string(self.pipe_id)
        bts.write_string(self.origin_node_id)
        bts.write_int(self.current_layer)
        bts.write_int(self.compute_step.value)
        bts.write_bytes(self.data.to_bytes() if self.data is not None else b'')
        bts.write_bytes(self.data_hash)

        bts.write_int(len(self.times))
        for time in self.times:
            bts.write_bytes(time.to_bytes())

        return bts.get_bytes()

    @staticmethod
    def from_bytes(data: bytes):
        bts = ByteHelper(data)

        job_id = bts.read_string()
        pipe_id = bts.read_string()
        origin_node_id = bts.read_string()
        current_layer = bts.read_int()
        step = ComputeStep(bts.read_int())
        job_bytes = bts.read_bytes()
        job_data = JobData.from_bytes(job_bytes) if job_bytes != b'' else None
        data_hash = bts.read_bytes()

        valid = True
        if data_hash != b'':
            valid = JobData.validate_state(job_bytes, data_hash)

        times = []
        for _ in range(0, bts.read_int()):
            times.append(JobTime.from_bytes(bts.read_bytes()))

        return NetworkJob(
            job_id=job_id, 
            pipe_id=pipe_id, 
            origin_node_id=origin_node_id, 
            current_layer=current_layer, 
            data=job_data, 
            data_hash=data_hash,
            compute_step=step,
            times=times
        ), valid



from uuid import uuid4
from logging import Logger
from typing import List, Optional, Tuple, Dict

from language_pipes.pipes.meta_pipe import MetaPipe
from language_pipes.pipes.router_pipes import RouterPipes

from language_pipes.modeling.llm_model import LlmModel
from language_pipes.modeling.end_model import EndModel
from language_pipes.modeling.llm_meta_data import validate_model

from language_pipes.config import LpConfig

class ModelManager:
    logger: Logger
    config: LpConfig
    router_pipes: RouterPipes
    models: List[LlmModel]
    end_models: List[EndModel]
    pipes_hosted: Dict[str, List[str]]

    def __init__(
        self, 
        logger: Logger,
        config: LpConfig,
        router_pipes: RouterPipes
    ):
        self.config = config
        self.logger = logger
        self.router_pipes = router_pipes
        self.models = []
        self.end_models = []
        self.pipes_hosted = { }
        for m in self.config.hosted_models:
            self._host_model(m.id, m.max_memory, m.device, m.load_ends)

    def stop(self):
        for m in self.models:
            m.cleanup_tensors()
        for m in self.end_models:
            m.clean_up()
        self.models = []
        self.end_models = []

    def get_end_model(self, model_id: str) -> Optional[EndModel]:
        for m in self.end_models:
            if m.model_id == model_id:
                return m
        return None

    def _get_model_for_pipe(self, model_id: str, pipe: MetaPipe, device: str, available_memory: int | float) -> Tuple[int | float, Optional[LlmModel]]:
        start_memory = available_memory

        new_model: Optional[LlmModel] = LlmModel.from_id(
            model_dir=self.config.model_dir,
            model_id=model_id,
            node_id=self.config.node_id,
            pipe_id=pipe.pipe_id,
            device=device
        )
        if new_model is None:
            return None
        meta_data = new_model.meta_data
        if self.config.model_validation and len(pipe.segments) > 0 and not validate_model(new_model.meta_data, pipe.get_computed()):
            self.logger.warning(f'Computed data for model {model_id} does not match')
            return available_memory, None
        
        num_layers_to_load = int(available_memory // meta_data.avg_layer_size) - 1
        total_layers = new_model.collector.config.num_hidden_layers
        start_layer = pipe.next_start_layer()
        if num_layers_to_load == -1:
            start_layer = -1
            end_layer = -1
        else:
            end_layer = min([start_layer + num_layers_to_load, pipe.next_end_layer(total_layers), new_model.num_hidden_layers]) if start_layer != -1 else -1
            available_memory = available_memory - (end_layer - start_layer + 1) * meta_data.avg_layer_size

        if num_layers_to_load > -1 and end_layer != -1 and start_layer != -1:
            self.logger.info(f'Using {(start_memory - available_memory) / 10**9:.2f} GB of memory to load model {model_id}')
            new_model.start_layer = start_layer
            new_model.end_layer = end_layer
            new_model.print(self.logger)
        else:
            new_model = None
        return available_memory, new_model

    def _load_end_model(self, model_id: str, device: str):
        model = EndModel(self.config.model_dir, model_id, device)
        self.end_models.append(model)
        return model

    def _host_model(self, model_id: str, max_memory: float, device: str, load_ends: bool):
        available_memory = max_memory * 10 ** 9
        models_to_load: List[LlmModel] = []
        end_model = None
        if load_ends:
            end_model = self._load_end_model(model_id, device)
        
        if model_id not in self.pipes_hosted:
            self.pipes_hosted[model_id] = []
        
        for pipe_id in [p.pipe_id for p in self.router_pipes.pipes_for_model(model_id, False)]:
            if pipe_id not in self.pipes_hosted[model_id] and len(self.pipes_hosted[model_id]) >= self.config.max_pipes:
                break
            loaded = True
            while loaded:
                pipe = self.router_pipes.get_pipe_by_pipe_id(pipe_id)
                if pipe is None: 
                    break
                available_memory, model = self._get_model_for_pipe(model_id, pipe, device, available_memory)
                loaded = model is not None
                if model is not None:
                    self.pipes_hosted[model_id].append(model.pipe_id)
                    self.router_pipes.add_model_to_network(model.to_meta())
                    models_to_load.append(model)

        if len(self.pipes_hosted[model_id]) < self.config.max_pipes:
            new_pipe = MetaPipe(str(uuid4()), model_id, [])
            self.pipes_hosted[model_id].append(new_pipe.pipe_id)
            _, model = self._get_model_for_pipe(model_id, new_pipe, device, available_memory)
            if model is not None:
                self.router_pipes.add_model_to_network(model.to_meta())
                models_to_load.append(model)

        if load_ends and end_model is not None:
            end_model.load()

        for m in models_to_load:
            m.load()
            self.router_pipes.update_model(m.to_meta())
            self.models.append(m)



import torch
import hashlib
from typing import Optional, Tuple
from language_pipes.util.byte_helper import ByteHelper
from llm_layer_collector.compute import LLmComputationState

from language_pipes.util import tensor_to_bytes, bytes_to_tensor, maybeTo

class JobData:
    cache_position: Optional[torch.Tensor] = None
    causal_mask: Optional[torch.Tensor] = None
    sliding_causal_mask: Optional[torch.Tensor] = None
    position_ids: Optional[torch.Tensor] = None
    position_embeddings: Optional[Tuple[torch.Tensor, torch.Tensor]] = None
    position_embeddings_local: Optional[Tuple[torch.Tensor, torch.Tensor]] = None
    position_embeddings_global: Optional[Tuple[torch.Tensor, torch.Tensor]] = None
    state: Optional[torch.Tensor] = None

    def hash_state(self):
        return hashlib.sha256(self.to_bytes()).digest()

    def to_bytes(self) -> bytes:
        state_bytes = tensor_to_bytes(self.state)
        cache_position_bytes = tensor_to_bytes(self.cache_position)
        causal_mask_bytes = tensor_to_bytes(self.causal_mask)
        sliding_causal_mask_bytes = tensor_to_bytes(self.sliding_causal_mask)
        position_ids_bytes = tensor_to_bytes(self.position_ids)
        position_embeddings_bytes = (
            (tensor_to_bytes(self.position_embeddings[0]), tensor_to_bytes(self.position_embeddings[1]))
            if self.position_embeddings is not None else (b'', b'')
        )
        position_embeddings_local_bytes = (
            (tensor_to_bytes(self.position_embeddings_local[0]), tensor_to_bytes(self.position_embeddings_local[1]))
            if self.position_embeddings_local is not None else (b'', b'')
        )
        position_embeddings_global_bytes = (
            (tensor_to_bytes(self.position_embeddings_global[0]), tensor_to_bytes(self.position_embeddings_global[1]))
            if self.position_embeddings_global is not None else (b'', b'')
        )

        bts = ByteHelper()

        bts.write_bytes(state_bytes)
        bts.write_bytes(cache_position_bytes)
        bts.write_bytes(causal_mask_bytes)
        bts.write_bytes(sliding_causal_mask_bytes)
        bts.write_bytes(position_ids_bytes)
        bts.write_bytes(position_embeddings_bytes[0])
        bts.write_bytes(position_embeddings_bytes[1])
        bts.write_bytes(position_embeddings_local_bytes[0])
        bts.write_bytes(position_embeddings_local_bytes[1])
        bts.write_bytes(position_embeddings_global_bytes[0])
        bts.write_bytes(position_embeddings_global_bytes[1])

        return bts.get_bytes()

    @staticmethod
    def from_bytes(data: bytes) -> Optional['JobData']:
        job_data = JobData()
        bts = ByteHelper(data)
        job_data.state = bytes_to_tensor(bts.read_bytes())
        job_data.cache_position = bytes_to_tensor(bts.read_bytes())
        job_data.causal_mask = bytes_to_tensor(bts.read_bytes())
        job_data.sliding_causal_mask = bytes_to_tensor(bts.read_bytes())
        job_data.position_ids = bytes_to_tensor(bts.read_bytes())
        pe0 = bytes_to_tensor(bts.read_bytes())
        pe1 = bytes_to_tensor(bts.read_bytes())
        job_data.position_embeddings = (pe0, pe1) if pe0 is not None else None
        
        pel0 = bytes_to_tensor(bts.read_bytes())
        pel1 = bytes_to_tensor(bts.read_bytes())
        job_data.position_embeddings_local = (pel0, pel1) if pel0 is not None else None
        
        peg0 = bytes_to_tensor(bts.read_bytes())
        peg1 = bytes_to_tensor(bts.read_bytes())
        job_data.position_embeddings_global = (peg0, peg1) if peg0 is not None else None
    
        return job_data

    @staticmethod
    def validate_state(data: bytes, state_hash: bytes) -> bool:
        current_hash = hashlib.sha256(data).digest()
        return current_hash == state_hash

def move_position_embeddings(t: Optional[Tuple[torch.Tensor, torch.Tensor]], device: str) -> Optional[Tuple[torch.Tensor, torch.Tensor]]:
    if t is None:
        return None
    if str(t[0].device) == device:
        return (t[0].detach(), t[1].detach())
    return (
        t[0].detach().to(device),
        t[1].detach().to(device)
    )

def computationStateToJobData(data: LLmComputationState) -> JobData:
    job_data = JobData()
    job_data.state = maybeTo(data.state, 'cpu')
    job_data.position_ids = maybeTo(data.position_ids, 'cpu')
    job_data.position_embeddings = move_position_embeddings(data.position_embeddings, 'cpu')
    job_data.position_embeddings_local = move_position_embeddings(data.position_embeddings_local, 'cpu')
    job_data.position_embeddings_global = move_position_embeddings(data.position_embeddings_global, 'cpu')
    job_data.cache_position = maybeTo(data.cache_position, 'cpu')
    job_data.causal_mask = maybeTo(data.causal_mask["full_attention"], 'cpu')
    job_data.sliding_causal_mask = maybeTo(data.causal_mask["sliding_attention"], 'cpu')
    return job_data

def jobDataToComputationState(data: JobData, device: str) -> LLmComputationState:
    state = LLmComputationState()
    state.state = maybeTo(data.state, device)
    state.position_ids = maybeTo(data.position_ids, device)
    
    state.position_embeddings = move_position_embeddings(data.position_embeddings, device)
    state.position_embeddings_local = move_position_embeddings(data.position_embeddings_local, device)
    state.position_embeddings_global = move_position_embeddings(data.position_embeddings_global, device)
    
    state.cache_position = maybeTo(data.cache_position, device)
    state.causal_mask = {
        "full_attention": maybeTo(data.causal_mask, device),
        "sliding_attention": maybeTo(data.sliding_causal_mask, device)
    }
    return state

def detachCompState(state: LLmComputationState) -> LLmComputationState:
    state.state = state.state.detach()
    state.position_ids = state.position_ids.detach()
    if state.position_embeddings is not None:
        state.position_embeddings = (state.position_embeddings[0].detach(), state.position_embeddings[1].detach())
    if state.position_embeddings_local is not None:
        state.position_embeddings_local = (state.position_embeddings_local[0].detach(), state.position_embeddings_local[1].detach())
    if state.position_embeddings_global is not None:
        state.position_embeddings_global = (state.position_embeddings_global[0].detach(), state.position_embeddings_global[1].detach())
    
    state.cache_position = state.cache_position.detach()
    state.causal_mask = {
        "full_attention": state.causal_mask["full_attention"].detach() if state.causal_mask["full_attention"] is not None else None,
        "sliding_attention": state.causal_mask["sliding_attention"].detach() if state.causal_mask["sliding_attention"] is not None else None
    }
    return state



import gc
import ctypes
import torch
from time import time
from typing import List, Optional
from time import time, sleep
from threading import Thread

from language_pipes.jobs.job import Job
from language_pipes.jobs.network_job import NetworkJob
from language_pipes.config import LpConfig

CHECK_JOB_INTERVAL = 10
EXPIRED_JOB_TIME = 60  # Unified timeout for both prefill and decode phases

try:
    _libc = ctypes.CDLL("libc.so.6")
    _malloc_trim = _libc.malloc_trim
    _malloc_trim.argtypes = [ctypes.c_size_t]
    _malloc_trim.restype = ctypes.c_int
except:
    _malloc_trim = None

class JobTracker:
    jobs_completed: List[str]
    jobs_pending: List[Job]

    def __init__(self, logger, config: LpConfig):
        self.logger = logger
        self.config = config
        self.jobs_completed = []
        self.jobs_pending = []
        Thread(target=self.check_stale_jobs, args=( )).start()

    def check_stale_jobs(self):
        while True:
            remove_jobs = []
            for j in self.jobs_pending:
                stale_time = time() - j.last_update
                # Unified timeout - prefill chunks regularly update last_update,
                # so both prefill and decode phases use the same timeout
                if stale_time > EXPIRED_JOB_TIME:
                    self.logger.warning(
                        f"[Stale] job={j.job_id[:8]} timed out after {stale_time:.1f}s "
                        f"(token={j.current_token})"
                    )
                    remove_jobs.append(j.job_id)

            if len(remove_jobs) == 0:
                sleep(CHECK_JOB_INTERVAL)
                continue
        
            for job_id in remove_jobs:
                self.jobs_pending = [j for j in self.jobs_pending if j.job_id != job_id]
            
            gc.collect()
            torch.cuda.empty_cache()
            if _malloc_trim is not None:
                _malloc_trim(0)

            sleep(CHECK_JOB_INTERVAL)

    def get_job(self, job_id: str) -> Optional[Job]:
        for j in self.jobs_pending:
            if j.job_id == job_id:
                return j
        return None

    def complete_job(self, job: Job):
        job_id = job.job_id
        if job_id in self.jobs_completed:
            return
        self.jobs_completed.append(job_id)
        if job.resolve is None:
            return
        self.logger.info(f'Received job complete for {job_id}\n')
        self.logger.info("[TIMING] Prefill Summary:")
        job.timing_stats.prefill_times.log_summary(self.logger)
        self.logger.info("[TIMING] Output Summary:")
        job.timing_stats.output_times.log_summary(self.logger)
        job.resolve(job)
        self.jobs_pending = [j for j in self.jobs_pending if j.job_id != job_id]

    def update_job_time(self, job_id: str):
        """Update the last_update time for a pending job to prevent stale timeout."""
        job = self.get_job(job_id)
        if job is None:
            return
        job.last_update = time()

    def add_job(self, network_job: NetworkJob) -> Job | None:
        existing = self.get_job(network_job.job_id)
        if existing is not None:
            return None
        
        job = Job(
            origin_node_id=network_job.origin_node_id,
            messages=[],
            model_id="",
            prefill_chunk_size=self.config.prefill_chunk_size,
            pipe_id=network_job.pipe_id,
            data=network_job.data
        )
        job.job_id = network_job.job_id
        
        if network_job.data.state is None:
            raise Exception("job should be embedded before adding a pending job")
        
        job.prompt_tokens = network_job.data.state.size()[1]
        job.last_update = time()
        self.jobs_pending.append(job)
        return job



from typing import Dict, List, Optional
from dataclasses import dataclass
from pathlib import Path

def default_config_dir() -> str:
    return str(Path.home() / ".config" / "language_pipes")


def default_model_dir() -> str:
    return str(Path.home() / ".cache" / "language_pipes" / "models")


@dataclass
class HostedModel:
    id: str
    device: str
    max_memory: float  # in gigabytes
    load_ends: bool  # Loads head and embed of model

    @staticmethod
    def from_dict(data: Dict) -> 'HostedModel':
        return HostedModel(
            id=data['id'],
            device=data['device'],
            max_memory=float(data['max_memory']),
            load_ends=data.get('load_ends', False)
        )

@dataclass
class LpConfig:
    # Core settings
    node_id: str
    app_dir: str
    model_dir: str
    logging_level: str
    
    # API server
    oai_port: Optional[int]
    
    # Model hosting
    hosted_models: List[HostedModel]
    
    # Processing options
    max_pipes: int
    model_validation: bool
    prefill_chunk_size: int

    @staticmethod
    def from_dict(data: Dict) -> 'LpConfig':
        if data.get("node_id") is None:
            raise Exception("Node ID must be supplied to config")
        if data.get("hosted_models") is None:
            raise Exception("Hosted models must be supplied to the configuration")
        
        logging_level = data.get('logging_level', None)
        if logging_level is None:
            logging_level = "INFO"
        
        app_dir = data.get('app_dir', None)
        if app_dir is None:
            app_dir = default_config_dir()
        
        model_dir = data.get('model_dir', None)
        if model_dir is None:
            model_dir = default_model_dir()
        
        model_validation = data.get('model_validation', None)
        if model_validation is None:
            model_validation = False
        
        prefill_chunk_size = data.get('prefill_chunk_size', None)
        if prefill_chunk_size is None:
            prefill_chunk_size = 6
        
        max_pipes = data.get('max_pipes', None)
        if max_pipes is None:
            max_pipes = 1

        return LpConfig(
            # Core settings
            node_id=data.get('node_id'),
            logging_level=logging_level,
            app_dir=app_dir,
            model_dir=model_dir,
            
            # API server
            oai_port=data.get('oai_port', None),
            
            # Model hosting
            hosted_models=[HostedModel.from_dict(m) for m in data['hosted_models']],
            
            # Processing options
            max_pipes=max_pipes,
            model_validation=model_validation,
            prefill_chunk_size=prefill_chunk_size
        )

    def to_string(self) -> str:
        lines = []
        
        lines.append("")
        lines.append("=" * 60)
        lines.append("  Configuration Details")
        lines.append("=" * 60)
        
        # Core settings
        lines.append("")
        lines.append("--- Core Settings ---")
        lines.append(f"  {'Node ID:':<18} {self.node_id}")
        lines.append(f"  {'App Directory:':<18} {self.app_dir}")
        lines.append(f"  {'Model Directory:':<18} {self.model_dir}")
        lines.append(f"  {'Logging Level:':<18} {self.logging_level}")
        
        # API settings
        lines.append("")
        lines.append("--- API Settings ---")
        if self.oai_port:
            lines.append(f"  {'OpenAI API Port:':<18} {self.oai_port}")
        else:
            lines.append("  OpenAI API:         Disabled")
        
        # Processing options
        lines.append("")
        lines.append("--- Processing Options ---")
        lines.append(f"  {'Max Pipes:':<18} {self.max_pipes}")
        lines.append(f"  {'Model Validation:':<18} {'Enabled' if self.model_validation else 'Disabled'}")
        lines.append(f"  {'Prefill Chunk Size:':<18} {self.prefill_chunk_size}")
        
        # Hosted models
        lines.append("")
        lines.append(f"--- Hosted Models ({len(self.hosted_models)}) ---")
        for i, model in enumerate(self.hosted_models):
            lines.append("")
            lines.append(f"  Model #{i+1}:")
            lines.append(f"    ID:          {model.id}")
            lines.append(f"    Device:      {model.device}")
            lines.append(f"    Max Memory:  {model.max_memory} GB")
            lines.append(f"    Load Ends:   {'Yes' if model.load_ends else 'No'}")
        
        lines.append("")
        lines.append("=" * 60)
        
        return "\n".join(lines)



from typing import List
from dataclasses import dataclass
from language_pipes.modeling.meta_model import MetaModel
from language_pipes.modeling.llm_meta_data import LlmMetadata

@dataclass
class MetaPipe:
    pipe_id: str
    model_id: str

    segments: List[MetaModel]

    def num_layers(self):
        if len(self.segments) == 0:
            return 0
        else:
            return self.segments[0].num_layers

    def is_loading(self) -> bool:
        return len([s for s in self.segments if not s.loaded]) > 0

    def get_computed(self) -> LlmMetadata:
        return self.segments[0].meta_data

    def sort_segments(self):
        self.segments = sorted(self.segments, key=lambda x: x.start_layer)

    def get_filled_slots(self, num_layers: int | None = None):
        if num_layers is None:
            num_layers = self.num_layers()
        filled_slots = [0 for _ in range(0, num_layers)]
        for segment in self.segments:
            if segment.start_layer == -1:
                continue
            for i in range(segment.start_layer, min([segment.end_layer + 1, num_layers])):
                filled_slots[i] = 2 if segment.loaded else 1
        return filled_slots

    def print_pipe(self):
        filled_slots = self.get_filled_slots()
        num_layers = len(filled_slots)
        pipe_pieces = ["|>"]
        for slot in filled_slots:
            match (slot):
                case 0:
                    pipe_pieces.append("X")
                case 1:
                    pipe_pieces.append("|")
                case 2:
                    pipe_pieces.append("=")
        pipe_pieces.append("<|")
        return "".join(pipe_pieces)

    def next_start_layer(self) -> int:
        if len(self.segments) == 0:
            return 0
        filled_slots = self.get_filled_slots()
        for slot in range(0, self.num_layers()):
            if filled_slots[slot] == 0:
                return slot
        return -1

    def next_end_layer(self, num_layers: int | None = None) -> int:
        if num_layers is None:
            num_layers = self.num_layers()
        start = self.next_start_layer()
        filled_slots = self.get_filled_slots(num_layers)
        for end_layer in range(start, num_layers):
            if end_layer == num_layers - 1:
                return end_layer
            if filled_slots[end_layer] > 0:
                return end_layer - 1
        return -1

    def peers(self) -> List[str]:
        peers: List[str] = []
        for segment in self.segments:
            if segment.node_id not in peers:
                peers.append(segment.node_id)
        return peers

    def is_complete(self):
        self.sort_segments()
        current_layer = 0
        for s in self.segments:
            if s.start_layer == -1 or not s.loaded:
                continue
            if s.start_layer == current_layer:
                current_layer = s.end_layer + 1

        return current_layer == self.segments[0].num_layers

    def print(self, logger):
        self.sort_segments()
        logger.info(f'''
=================================
Pipe Status:
Model ID: {self.model_id}
Pipe: {self.pipe_id}
Segments: {', '.join([s.node_id for s in self.segments])}
{self.print_pipe()}
End Layer: {self.segments[-1].end_layer} / {self.num_layers() - 1}
Complete: {self.is_complete()}
=================================
''')



import os
import sys
import torch
import unittest

sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', '..', '..', '..', 'src'))
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', '..', '..', '..', 'tests', 'language_pipes', 'unit'))

from language_pipes.jobs.job_data import JobData
from language_pipes.jobs.job_processor import JobState
from language_pipes.util.enums import ComputeStep

from job_processor.util import make_processor, make_job, make_config, FakeModel, FakePipe

class TestSendState(unittest.TestCase):
    """Tests for the _state_send method."""

    def test_transitions_to_done_after_handoff(self):
        job = make_job()
        job.compute_step = ComputeStep.LAYER
        job.current_layer = 0
        job.data = JobData()
        job.data.state = torch.zeros((1, 1))

        pipe = FakePipe(FakeModel("node-b", 0, 0, virtual=True))
        processor = make_processor(job=job, pipe=pipe, end_model=None)

        next_state = processor._state_send()

        self.assertEqual(next_state, JobState.DONE)
        self.assertEqual(len(pipe.sent_jobs), 1)
        _, node_id = pipe.sent_jobs[0]
        self.assertEqual(node_id, "node-b")

    def test_routes_tokenize_job_to_next_node_when_origin_mismatch(self):
        job = make_job(origin_node_id="node-b")
        job.compute_step = ComputeStep.TOKENIZE
        job.current_layer = 0
        job.data = JobData()
        job.data.state = torch.zeros((1, 1))

        pipe = FakePipe(FakeModel("node-c", 0, 0))
        processor = make_processor(
            job=job,
            pipe=pipe,
            config=make_config(node_id="node-a"),
            end_model=None,
        )

        processor.run()

        self.assertEqual(len(pipe.sent_jobs), 1)
        _, node_id = pipe.sent_jobs[0]
        self.assertEqual(node_id, "node-c")



from typing import Optional

from language_pipes.config import LpConfig

from language_pipes.pipes.pipe import Pipe
from language_pipes.pipes.meta_pipe import MetaPipe
from language_pipes.pipes.router_pipes import RouterPipes

from language_pipes.modeling.model_manager import ModelManager

class PipeManager:
    config: LpConfig
    router_pipes: RouterPipes
    model_manager: ModelManager

    def __init__(
        self,
        config: LpConfig,
        model_manager: ModelManager,
        router_pipes: RouterPipes
    ):
        self.config = config
        self.model_manager = model_manager
        self.router_pipes = router_pipes

    def _get_pipe_from_meta(self, meta_pipe: MetaPipe) -> Pipe:
        return Pipe.from_meta(
            meta_pipe=meta_pipe,
            hosted_models=self.model_manager.models,
            router=self.router_pipes.router,
            model_dir=self.config.model_dir
        )

    def get_pipe_by_pipe_id(self, pipe_id: str) -> Optional[Pipe]:
        meta_pipe = self.router_pipes.get_pipe_by_pipe_id(pipe_id)
        if meta_pipe is None:
            return None
        return self._get_pipe_from_meta(meta_pipe)

    def get_pipe_by_model_id(self, model_id: str) -> Optional[Pipe]:
        meta_pipe = self.router_pipes.get_pipe_by_model_id(model_id)
        if meta_pipe is None:
            return None
        return self._get_pipe_from_meta(meta_pipe)



import random
from time import sleep
from threading import Thread
from typing import Callable, Optional, List

from language_pipes.pipes.pipe_manager import PipeManager

from language_pipes.jobs.job import ComputeStep
from language_pipes.jobs.job_factory import JobFactory
from language_pipes.jobs.job_tracker import JobTracker
from language_pipes.jobs.network_job import NetworkJob
from language_pipes.jobs.job_processor import JobProcessor, JobContext

from language_pipes.modeling.model_manager import ModelManager

from language_pipes.config import LpConfig

class JobReceiver:
    config: LpConfig
    job_factory: JobFactory
    job_queue: List[NetworkJob]
    pipe_manager: PipeManager
    model_manager: ModelManager
    is_shutdown: Callable[[], bool]

    def __init__(
            self, 
            config: LpConfig,
            logger,
            job_factory: JobFactory,
            job_tracker: JobTracker,
            pipe_manager: PipeManager,
            model_manager: ModelManager,
            is_shutdown: Callable[[], bool]
    ):
        self.job_queue = []
        self.job_tracker = job_tracker
        self.job_factory = job_factory
        self.model_manager = model_manager
        self.pipe_manager = pipe_manager
        self.config = config
        self.is_shutdown = is_shutdown
        self.logger = logger

        Thread(target=self._job_runner_loop, args=()).start()

    def _wait_for_job(self) -> Optional[NetworkJob]:
        """Wait for a job from the queue. Returns None if shutting down."""
        while True:
            if self.is_shutdown():
                return None
            if len(self.job_queue) > 0:
                idx = random.randrange(len(self.job_queue))
                network_job = self.job_queue.pop(idx)
                return network_job
            sleep(0.01)

    def _job_runner_loop(self):
        """Main job processing loop using FSM."""
        network_job = self._wait_for_job()
        if network_job is None:
            return
        
        job = self.job_tracker.get_job(network_job.job_id)
        if job is None:
            job = self.job_tracker.add_job(network_job)

        # Validate network job
        if not job.receive_network_job(network_job):
            return

        pipe = self.pipe_manager.get_pipe_by_pipe_id(network_job.pipe_id)
        if pipe is None:
            return

        end_model = self.model_manager.get_end_model(pipe.model_id)
        
        fsm = JobProcessor(JobContext(
            logger=self.logger,
            pipe=pipe,
            end_model=end_model,
            job=job,
            config=self.config
        ))

        try:
            fsm.run()
        except Exception as e:
            print(e)
        
        Thread(target=self._job_runner_loop, args=()).start()

    def restart_token(self, network_job: NetworkJob):
        """Mark job for restart and send back to origin."""
        network_job.data = None
        network_job.data_hash = b''
        network_job.compute_step = ComputeStep.EMBED
        network_job.current_layer = 0
        pipe = self.pipe_manager.get_pipe_by_pipe_id(network_job.pipe_id)
        if pipe is None:
            return
        pipe.send_job(network_job, network_job.origin_node_id)

    def receive_data(self, data: bytes):
        """Receive and validate incoming job data."""
        job, valid = NetworkJob.from_bytes(data)
        if not valid:
            self.restart_job(job)
            return
        
        # Ignore duplicate jobs
        for j in self.job_queue:
            if j.job_id == job.job_id:
                return

        self.job_queue.insert(0, job)



from distributed_state_network.objects.signed_packet import SignedPacket
from distributed_state_network.util.byte_helper import ByteHelper

class DataPacket(SignedPacket):
    node_id: str
    data: bytes

    def __init__(self, node_id: str, ecdsa_signature: bytes, data: bytes):
        super().__init__(ecdsa_signature)
        self.node_id = node_id
        self.data = data

    def to_bytes(self, include_signature: bool = True) -> bytes:
        bts = ByteHelper()
        bts.write_string(self.node_id)
        if include_signature:
            bts.write_bytes(self.ecdsa_signature)
        bts.write_bytes(self.data)
        return bts.get_bytes()

    @staticmethod
    def from_bytes(data: bytes) -> "DataPacket":
        bts = ByteHelper(data)
        node_id = bts.read_string()
        if node_id == "":
            raise Exception(406, "Malformed packet")
        ecdsa_signature = bts.read_bytes()
        payload = bts.read_bytes()
        return DataPacket(node_id, ecdsa_signature, payload)

    @staticmethod
    def create(node_id: str, ecdsa_private_key: bytes, data: bytes) -> "DataPacket":
        pkt = DataPacket(node_id, b"", data)
        pkt.sign(ecdsa_private_key)
        return pkt



from enum import Enum, auto
from dataclasses import dataclass
from logging import Logger

from language_pipes.jobs.job import Job

from language_pipes.pipes.pipe import Pipe
from language_pipes.modeling.end_model import EndModel

from language_pipes.util.enums import ComputeStep, JobStatus
from language_pipes.config import LpConfig

class JobState(Enum):
    VALIDATING = auto()    # Validating pipe and getting resources
    HEAD = auto()          # Computing norm/head, handling completion
    EMBED = auto()         # Embedding the next token for decoding
    PROCESS_LAYERS = auto() # Processing through local layers
    SEND = auto()          # Sending job to next destination
    DONE = auto()          # Current job iteration complete

@dataclass
class JobContext:
    job: Job
    pipe: Pipe
    logger: Logger
    config: LpConfig
    end_model: EndModel

def should_prefill_chunk(job) -> bool:
    return job.current_token == 0 and job.chunking.has_more()

def log_done_error(ctx: JobContext, message: str) -> None:
    if ctx.logger is not None:
        ctx.logger.error(message)

def get_next_state(ctx: JobContext) -> JobState:
    cs = ctx.job.compute_step
    if cs == ComputeStep.HEAD or cs == ComputeStep.EMBED or cs == ComputeStep.TOKENIZE:
        if ctx.job.origin_node_id != ctx.config.node_id:
            return JobState.SEND
        
        if should_prefill_chunk(ctx.job) or cs == ComputeStep.EMBED or cs == ComputeStep.TOKENIZE:
            return JobState.EMBED  
        else:
            return JobState.HEAD

    model = ctx.pipe.get_layer(ctx.job.current_layer, False)
    if model is None:
        log_done_error(
            ctx,
            f"[FSM] Missing model for layer={ctx.job.current_layer}; completing with error."
        )
        return JobState.DONE

    if model.virtual:
        return JobState.SEND
    return JobState.PROCESS_LAYERS

class JobProcessor:
    """
    Finite state machine for processing jobs.
    
    State Transitions:
    
    VALIDATING -> DONE (missing job/context resources or pipe unavailable)
    VALIDATING -> HEAD (job.done and prefill finished or decode)
    VALIDATING -> EMBED (job.done and more prefill chunks)
    VALIDATING -> PROCESS_LAYERS (job still needs local layer processing)
    
    HEAD -> DONE (job complete or failed to send update)
    HEAD -> EMBED (more tokens to generate locally)
    HEAD -> SEND (next layer is virtual/remote)
    HEAD -> PROCESS_LAYERS (next layer is local)

    EMBED -> DONE (failed to send update or missing model)
    EMBED -> SEND (next layer is virtual/remote)
    EMBED -> PROCESS_LAYERS (next layer is local)
    
    PROCESS_LAYERS -> DONE (missing local model)
    PROCESS_LAYERS -> SEND (next layer set is not local)
    PROCESS_LAYERS -> PROCESS_LAYERS (next layer set is local)
    
    SEND -> DONE (handoff complete)
    """
    
    state: JobState
    ctx: JobContext
    
    def __init__(self, ctx: JobContext):
        self.state = JobState.VALIDATING
        self.ctx = ctx
    
    def run(self):
        while self.state != JobState.DONE:
            self.state = self._transition()
    
    def _transition(self) -> JobState:
        """Execute current state and transition to next."""
        match self.state:
            case JobState.VALIDATING:
                return self._state_validating()
            case JobState.HEAD:
                return self._state_head()
            case JobState.EMBED:
                return self._state_embed()
            case JobState.PROCESS_LAYERS:
                return self._state_process_layers()
            case JobState.SEND:
                return self._state_send()

        return JobState.DONE
    
    def _state_validating(self) -> JobState:
        """Validate context for processing"""
        if self.ctx.job is None:
            log_done_error(self.ctx, "[FSM] Missing job; completing with error.")
            return JobState.DONE
        
        pipe = self.ctx.pipe
        
        # Ensure we have an available pipe
        if pipe is None or not pipe.is_complete():
            log_done_error(self.ctx, "[FSM] Pipe unavailable or incomplete; completing with error.")
            return JobState.DONE
        
        if self.ctx.job.compute_step == ComputeStep.HEAD:
            # Ensure we only process the ends of jobs we sent out
            if self.ctx.job.origin_node_id != self.ctx.config.node_id:
                log_done_error(self.ctx, "[FSM] Layer job origin mismatch; completing with error.")
                return JobState.DONE
            
            # Ensure we have the end model ready            
            if self.ctx.end_model is None:
                log_done_error(self.ctx, "[FSM] End model unavailable; completing with error.")
                return JobState.DONE
            
            # Job returned from network - check pending job
            if self.ctx.job is None:
                log_done_error(self.ctx, "[FSM] Job missing; completing with error.")
                return JobState.DONE

        return get_next_state(self.ctx)

    def _state_head(self) -> JobState:
        """Handle norm/head computation and prepare to embed the next token."""
        job = self.ctx.job
        end_model = self.ctx.end_model

        # Log prefill completion when transitioning from prefill to decode
        if job.current_token == 0:
            if job.chunking.has_more():
                log_done_error(self.ctx, "Received head state for job that was not done chunking")
                return JobState.DONE

            job.chunking.disable()
        
        job.compute_step = ComputeStep.NORM
        job.current_layer = 0

        job.timing_stats.add_head_time(self.ctx.config.node_id)
        end_model.compute_norm(job)
        end_model.compute_head(job)
        job.timing_stats.set_send_time(self.ctx.logger)
        job.timing_stats.finalize_token(self.ctx.logger)
        job.print_job(self.ctx.logger)

        # Job completed
        if job.status == JobStatus.COMPLETED:
            end_model.set_result(job)
            job.complete()
            return JobState.DONE
        
        # More tokens to generate - update and continue
        if not job.send_update():
            log_done_error(self.ctx, "[FSM] Failed to send job update; completing with error.")
            return JobState.DONE

        return JobState.EMBED

    def _state_embed(self) -> JobState:
        """Embed the next token, handling prefill chunks when needed."""
        job = self.ctx.job
        end_model = self.ctx.end_model
        logger = self.ctx.logger

        if job.prompt_tokens == 0:
            end_model.tokenize(job)
            logger.info(f"[Prefill] job={job.job_id[:8]} started")
            job.init_chunking(self.ctx.config.prefill_chunk_size)
            job.chunking.print_start(logger)
        elif job.chunking.is_active():
            job.chunking.advance()
            job.timing_stats.finalize_prefill_chunk(logger)
            logger.info(f"[PREFILL] Chunk {job.chunking.current_chunk} / {job.chunking.total_chunks} ({(job.chunking.current_chunk / job.chunking.total_chunks) * 100.0:.2f}%)")
            job.delta = ""
            if not job.send_update():
                log_done_error(self.ctx, "[FSM] Failed to send prefill job update; completing with error.")
                return JobState.DONE
        
        job.timing_stats.add_embed_time(self.ctx.config.node_id)
        end_model.compute_embed(job, logger, self.ctx.config.prefill_chunk_size)
        job.timing_stats.set_send_time(logger)
        
        return get_next_state(self.ctx)

    def _state_process_layers(self) -> JobState:
        """Process job through local layers."""
        pipe = self.ctx.pipe
        job = self.ctx.job
        
        model = pipe.get_layer(job.current_layer, True)
        if model is None:
            log_done_error(
                self.ctx,
                f"[FSM] Missing local model for layer={job.current_layer}; completing with error."
            )
            return JobState.DONE
        
        model.process_job(job, self.ctx.logger)
        job.set_last_update()
        
        return get_next_state(self.ctx)
    
    def _state_send(self) -> JobState:
        """Send job to next destination."""
        job = self.ctx.job
        pipe = self.ctx.pipe
        network_job = job.to_network_job()

        if job.compute_step == ComputeStep.HEAD:
            pipe.send_job(network_job, network_job.origin_node_id)
        else:
            next_model = pipe.get_layer(network_job.current_layer, False)
            if next_model is None:
                return JobState.DONE
            pipe.send_job(network_job, next_model.node_id)
        
        return JobState.DONE



import socket
import toml
from typing import List

from language_pipes.commands.view import view_config
from language_pipes.util.aes import generate_aes_key
from language_pipes.util.user_prompts import prompt, prompt_bool, prompt_choice, prompt_float, prompt_int, prompt_number_choice, show_banner, prompt_model_id

def edit_config(config_path: str):
    with open(config_path, 'r', encoding='utf-8') as f:
        config = toml.load(f)
    
    while True:
        show_banner("Edit Configuration")        
        view_config(config_path)
        
        editable_props = []
        prop_keys = []
        
        simple_props = [
            ("node_id", "Node ID"),
            ("oai_port", "OpenAI API Port"),
            ("peer_port", "Peer Port"),
            ("network_ip", "Network IP"),
            ("bootstrap_address", "Bootstrap Address"),
            ("bootstrap_port", "Bootstrap Port"),
            ("network_key", "Network Key"),
            ("logging_level", "Logging Level"),
            ("max_pipes", "Max Pipes"),
            ("model_validation", "Model Validation"),
            ("prefill_chunk_size", "Prefill Chunk Size"),
        ]
        
        model_count = len(config.get("hosted_models", []))
        editable_props.append(f"Hosted Models ({model_count} configured)")
        prop_keys.append("hosted_models")
        
        for key, label in simple_props:
            current_val = config.get(key, "Not set")
            editable_props.append(f"{label}: {current_val}")
            prop_keys.append(key)
        
        editable_props.append("Save and Exit")
        prop_keys.append("__save__")
        
        editable_props.append("Exit without Saving")
        prop_keys.append("__cancel__")
        
        selection = prompt_number_choice("\nSelect property to edit", editable_props, required=True)
        selected_key = prop_keys[editable_props.index(selection)]
        
        if selected_key == "__save__":
            with open(config_path, 'w', encoding='utf-8') as f:
                toml.dump(config, f)
            print("\n✓ Configuration saved")
            return
        
        if selected_key == "__cancel__":
            print("\nChanges discarded")
            return
        
        if selected_key == "node_id":
            config["node_id"] = prompt(
                "Node ID",
                default=config.get("node_id", socket.gethostname()),
                required=True
            )
        
        elif selected_key == "oai_port":
            current = config.get(selected_key)
            if prompt_bool("Enable OpenAI API?", default=current is not None):
                config[selected_key] = prompt_int(
                    "API Port",
                    default=current or 8000,
                    required=True
                )
            else:
                config.pop(selected_key, None)
        
        elif selected_key == "peer_port":
            config[selected_key] = prompt_int(
                "Peer Port",
                default=config.get(selected_key, 5000),
                required=True
            )
        
        elif selected_key == "network_ip":
            config[selected_key] = prompt(
                "Network IP",
                default=config.get(selected_key, None),
                required=True
            )
        
        elif selected_key == "bootstrap_address":
            current = config.get(selected_key)
            if prompt_bool("Connect to bootstrap node?", default=current is not None):
                config[selected_key] = prompt(
                    "Bootstrap Address",
                    default=current,
                    required=True
                )
            else:
                config.pop(selected_key, None)
        
        elif selected_key == "bootstrap_port":
            config[selected_key] = prompt_int(
                "Bootstrap Port",
                default=config.get(selected_key, 5000),
                required=True
            )
        
        elif selected_key == "network_key":
            current = config.get(selected_key)
            if prompt_bool("Enable network encryption?", default=current is not None):
                new_key = prompt(
                    "Network Key",
                    default=current or "Generate new key"
                )
                if new_key == "Generate new key":
                    key = generate_aes_key().hex()
                    config[selected_key] = key
                    print(f"Generated new key: {key}")
                else:
                    config[selected_key] = new_key
            else:
                config.pop(selected_key, None)
        
        elif selected_key == "logging_level":
            config[selected_key] = prompt_choice(
                "Logging Level",
                ["DEBUG", "INFO", "WARNING", "ERROR"],
                default=config.get(selected_key, "INFO")
            )
        
        elif selected_key == "max_pipes":
            config[selected_key] = prompt_int(
                "Max Pipes",
                default=config.get(selected_key, 1),
                required=True
            )
        
        elif selected_key == "model_validation":
            config[selected_key] = prompt_bool(
                "Enable model hash validation?",
                default=config.get(selected_key, False)
            )
        
        elif selected_key == "prefill_chunk_size":
            config[selected_key] = prompt_int(
                "Prefill Chunk Size",
                default=config.get(selected_key, 6),
                required=True
            )
        
        elif selected_key == "hosted_models":
            try:
                config["hosted_models"] = edit_hosted_models(config.get("hosted_models", []))
            except KeyboardInterrupt:
                continue

def select_hosted_model(models: List[dict]) -> int | None:
    options = []
    for i, model in enumerate(models):
        model_str = f"Model #{i+1}: {model.get('id', 'Unknown')} ({model.get('device', 'cpu')}, {model.get('max_memory', 4)}GB)"
        options.append(model_str)
    options.append("Cancel")
    model_selection = prompt_number_choice("Select Model", options)
    if model_selection == "Cancel":
        return None
    return options.index(model_selection)

def edit_hosted_models(models: List) -> List[dict]:
    while True:
        print("\n--- Hosted Models ---\n")
        
        options = [
            "Add model",
            "Edit model",
            "Delete Model",
            "Done editing models"
        ]
        
        selection = prompt_number_choice("Select model to edit", options, required=True)
        
        if selection == "Add model":
            new_model = {
                "id": prompt_model_id("Model ID", True),
                "device": prompt("Device", "cpu", required=True),
                "max_memory": prompt("Max Memory (GB)"),
                "load_ends": prompt_bool("Load Ends", required=True)
            }
            if new_model:
                models.append(new_model)
            continue
        
        if selection == "Edit model":
            model_idx = select_hosted_model(models)
            if model_idx is None:
                continue
            try:
                edited = edit_single_model(models[model_idx])
                if edited:
                    models[model_idx] = edited
            except KeyboardInterrupt:
                continue

        if selection == "Delete model":
            model_idx = select_hosted_model(models)
            if model_idx is None:
                continue
            models.pop(model_idx)

        if selection == "Done editing models":
            return models

def edit_single_model(model: dict) -> dict | None:
    print("\n--- Edit Model ---\n")
    
    props = [
        ("id", "Model ID"),
        ("device", "Device"),
        ("max_memory", "Max Memory (GB)"),
        ("load_ends", "Load Ends")
    ]
    
    edited_model = model.copy()
    
    while True:
        options = [f"{label}: {edited_model.get(key, 'N/A')}" for key, label in props]
        options.append("Cancel")
        options.append("Done")
        
        selection = prompt_number_choice("Select property", options, required=True)
        if selection == "Cancel":
            return None
        
        if selection == "Done":
            return edited_model

        selected_idx = options.index(selection)
        selected_key = props[selected_idx][0]
        
        if selected_key == "id":
            edited_model[selected_key] = prompt_model_id("Model ID", required=True)
        
        elif selected_key == "device":
            edited_model[selected_key] = prompt(
                "Device (cpu, cuda:0, etc.)",
                default=edited_model.get(selected_key, "cpu"),
                required=True
            )
        
        elif selected_key == "max_memory":
            edited_model[selected_key] = prompt_float(
                "Max Memory (GB)",
                required=True
            )
        
        elif selected_key == "load_ends":
            edited_model["load_ends"] = prompt_bool(
                "Load embedding/output layers?",
                required=True
            )



import os
import sys
import unittest

sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', '..', 'src'))

from language_pipes.jobs.network_job import NetworkJob, JobTime
from language_pipes.util.enums import ComputeStep


class NetworkJobTests(unittest.TestCase):
    def test_layer_time_round_trip(self):
        layer_time = JobTime(
            node_id="node-a",
            is_embed=True,
            is_head=False,
            start_layer=0,
            end_layer=3,
        )
        layer_time.set_send_time()

        restored = JobTime.from_bytes(layer_time.to_bytes())

        self.assertEqual(restored.node_id, "node-a")
        self.assertTrue(restored.is_embed)
        self.assertFalse(restored.is_head)
        self.assertEqual(restored.start_layer, 0)
        self.assertEqual(restored.end_layer, 3)

    def test_network_job_round_trip(self):
        layer_time = JobTime(node_id="node-a", start_layer=0, end_layer=1)
        layer_time.set_send_time()
        job = NetworkJob(
            job_id="job-1",
            pipe_id="pipe-1",
            origin_node_id="node-a",
            current_layer=2,
            data=None,
            data_hash=b"",
            compute_step=ComputeStep.LAYER,
            times=[layer_time],
        )

        restored, _ = NetworkJob.from_bytes(job.to_bytes())

        self.assertEqual(restored.job_id, "job-1")
        self.assertEqual(restored.pipe_id, "pipe-1")
        self.assertEqual(restored.origin_node_id, "node-a")
        self.assertEqual(restored.current_layer, 2)
        self.assertEqual(restored.compute_step, ComputeStep.LAYER)
        self.assertEqual(len(restored.times), 1)
        self.assertEqual(restored.times[0].node_id, "node-a")


if __name__ == "__main__":
    unittest.main()



import sys
import threading
import logging
from typing import Callable, List, Optional
from flask import Flask, request, Response
from distributed_state_network.dsnode import DSNode
from distributed_state_network.objects.config import DSNodeConfig
from distributed_state_network.util.aes import generate_aes_key
from distributed_state_network.util import stop_thread

VERSION = "0.6.7"
logging.basicConfig(level=logging.INFO)

# Silence Flask and Werkzeug logging
logging.getLogger('werkzeug').setLevel(logging.ERROR)

# Message type constants
MSG_HELLO = 1
MSG_PEERS = 2
MSG_UPDATE = 3
MSG_PING = 4
MSG_DATA = 5

class DSNodeServer:
    config: DSNodeConfig
    running: bool
    app: Flask
    node: DSNode
    thread: Optional[threading.Thread]

    def __init__(
        self, 
        config: DSNodeConfig,
        disconnect_callback: Optional[Callable] = None,
        update_callback: Optional[Callable] = None,
        receive_callback: Optional[Callable] = None
    ):
        self.config = config
        self.running = False
        self.thread = None
        
        # Create Flask app
        self.app = Flask(__name__)
        self.app.logger.setLevel(logging.ERROR)  # Reduce Flask's logging noise
        
        # Create DSNode
        self.node = DSNode(config, VERSION, disconnect_callback, update_callback, receive_callback)
        
        # Set up Flask routes
        self._setup_routes()

    def _setup_routes(self):
        """Set up Flask routes for each message type"""
        
        @self.app.route('/hello', methods=['POST'])
        def handle_hello_route():
            return self._handle_request(MSG_HELLO, request.data, request.remote_addr)
        
        @self.app.route('/peers', methods=['POST'])
        def handle_peers_route():
            return self._handle_request(MSG_PEERS, request.data, request.remote_addr)
        
        @self.app.route('/update', methods=['POST'])
        def handle_update_route():
            return self._handle_request(MSG_UPDATE, request.data, request.remote_addr)
        
        @self.app.route('/ping', methods=['POST'])
        def handle_ping_route():
            return self._handle_request(MSG_PING, request.data, request.remote_addr)

        @self.app.route('/data', methods=['POST'])
        def handle_data_route():
            return self._handle_request(MSG_DATA, request.data, request.remote_addr)


    def _handle_request(self, msg_type: int, data: bytes, remote_addr: Optional[str]) -> Response:
        if not self.running:
            return Response(status=500)
        try:
            # Decrypt the data
            if self.config.aes_key is not None:
                data = self.node.decrypt_data(data)
            
            if len(data) < 1:
                return Response(status=400)
            
            # First byte should be message type (for verification)
            received_msg_type = data[0]
            body = data[1:]
            
            if received_msg_type != msg_type:
                self.node.logger.error(f"Message type mismatch: expected {msg_type}, got {received_msg_type}")
                return Response(status=400)
            
            response_data = None
            
            if msg_type == MSG_HELLO:
                # Pass the detected IP address to handle_hello
                if remote_addr is None:
                    raise ValueError("Must supply remote address with hello")
                response_data = self.node.handle_hello(body, remote_addr)
                
            elif msg_type == MSG_PEERS:
                response_data = self.node.handle_peers(body)
                
            elif msg_type == MSG_UPDATE:
                response_data = self.node.handle_update(body)
                
            elif msg_type == MSG_PING:
                response_data = b''

            elif msg_type == MSG_DATA:
                response_data = self.node.receive_data(body)
            
            # Send response if handler returned data
            if response_data is not None:
                # Prepend message type to response
                response_with_type = bytes([msg_type]) + response_data
                if self.config.aes_key is not None:
                    response_with_type = self.node.encrypt_data(response_with_type)
                return Response(response_with_type, status=200, mimetype='application/octet-stream')
            else:
                return Response(status=204)  # No content
                
        except Exception as e:
            if len(e.args) >= 2 and isinstance(e.args[0], int):
                # Error with HTTP status code
                self.node.logger.error(f"Error handling {msg_type} from {remote_addr}: {e.args[1]}")
                return Response(status=e.args[0])
            else:
                self.node.logger.error(f"Error handling {msg_type} from {remote_addr}: {e}")
                return Response(status=500)

    def stop(self):
        self.node.shutting_down = True
        self.running = False
        if self.thread is not None:
            stop_thread(self.thread)

    def _serve_forever(self, port: int):
        if self.running:
            return
        # Suppress Flask startup messages
        cli = sys.modules['flask.cli']
        cli.show_server_banner = lambda *x: None
        
        self.running = True
        self.app.run(host='0.0.0.0', port=port, threaded=True, use_reloader=False)
        self.node.logger.info(f'Started DSNode on HTTP port {port}')

    @staticmethod
    def generate_key() -> str:
        return generate_aes_key().hex()

    @staticmethod 
    def start(
        config: DSNodeConfig, 
        disconnect_callback: Optional[Callable] = None, 
        update_callback: Optional[Callable] = None,
        receive_callback: Optional[Callable] = None
    ) -> 'DSNodeServer':
        n = DSNodeServer(config, disconnect_callback, update_callback, receive_callback)
        n.thread = threading.Thread(target=n._serve_forever, daemon=True, args=(config.port, ))
        n.thread.start()

        if n.config.bootstrap_nodes is not None and len(n.config.bootstrap_nodes) > 0:
            for bs in n.config.bootstrap_nodes:
                try:
                    n.node.bootstrap(bs)
                    break # Throws exception if connection is not made
                except Exception as e:
                    print(e)

        return n

    def peers(self) -> List[str]:
        return self.node.peers()
    
    def read_data(self, node_id: str, key: str) -> Optional[str]:
        return self.node.read_data(node_id, key)
    
    def update_data(self, key: str, val: str):
        return self.node.update_data(key, val)

    def send_to_node(self, node_id: str, data: bytes):
        return self.node.send_to_node(node_id, data)

    def is_shut_down(self) -> bool:
        return self.node.shutting_down
    
    def node_id(self) -> str:
        return self.config.node_id

    def set_receive_cb(self, cb: Callable):
        self.node.receive_cb = cb

    def set_update_cb(self, cb: Callable):
        self.node.update_cb = cb

    def set_disconnect_cb(self, cb: Callable):
        self.node.disconnect_cb = cb

    def _receive_data(self, data: bytes):
        if self.node.receive_cb is not None:
            self.node.receive_cb(data)
